import itertools
import toolviper.utils.logger as logger

import numpy as np

try:
    from casacore import tables
except ImportError:
    import xradio._utils._casacore.casacore_from_casatools as tables

from ._tables.read import table_exists


def enumerated_product(*args):
    yield from zip(
        itertools.product(*(range(len(x)) for x in args)), itertools.product(*args)
    )


def create_partitions(in_file: str, partition_scheme: list) -> list[dict]:
    """Create a list of dictionaries with the partition information.

    Parameters
    ----------
    in_file: str
        Input MSv2 file path.
    partition_scheme:  list
        A MS v4 can only contain a single data description (spectral window and polarization setup), and observation mode. Consequently, the MS v2 is partitioned when converting to MS v4.
        In addition to data description and polarization setup a finer partitioning is possible by specifying a list of partitioning keys. Any combination of the following keys are possible:
        "FIELD_ID", "SCAN_NUMBER", "STATE_ID", "SOURCE_ID", "SUB_SCAN_NUMBER", "ANTENNA1".
        For mosaics where the phase center is rapidly changing (such as VLA on the fly mosaics)  partition_scheme should be set to an empty list []. By default, ["FIELD_ID"].
    Returns
    -------
    list
        list of dictionaries with the partition information.
    """
    # vla_otf (bool, optional):  The partioning of VLA OTF (on the fly) mosaics needs a special partitioning scheme. Defaults to False.

    # Create partition table
    import pandas as pd
    import os

    partition_scheme = ["DATA_DESC_ID", "OBS_MODE", "OBSERVATION_ID"] + partition_scheme

    # Open MSv2 tables and add columns to partition table (par_df):
    par_df = pd.DataFrame()
    main_tb = tables.table(
        in_file, readonly=True, lockoptions={"option": "usernoread"}, ack=False
    )
    par_df["DATA_DESC_ID"] = main_tb.getcol("DATA_DESC_ID")
    par_df["FIELD_ID"] = main_tb.getcol("FIELD_ID")
    par_df["SCAN_NUMBER"] = main_tb.getcol("SCAN_NUMBER")
    par_df["STATE_ID"] = main_tb.getcol("STATE_ID")
    par_df["OBSERVATION_ID"] = main_tb.getcol("OBSERVATION_ID")
    par_df["ANTENNA1"] = main_tb.getcol("ANTENNA1")
    par_df = par_df.drop_duplicates()

    field_tb = tables.table(
        os.path.join(in_file, "FIELD"),
        readonly=True,
        lockoptions={"option": "usernoread"},
        ack=False,
    )
    # if vla_otf:
    #     par_df["FIELD_NAME"] = np.array(field_tb.getcol("NAME"))[par_df["FIELD_ID"]]

    # Get source ids if available from source table.
    if table_exists(os.path.join(os.path.join(in_file, "SOURCE"))):
        source_tb = tables.table(
            os.path.join(in_file, "SOURCE"),
            readonly=True,
            lockoptions={"option": "usernoread"},
            ack=False,
        )
        if source_tb.nrows() != 0:
            par_df["SOURCE_ID"] = field_tb.getcol("SOURCE_ID")[par_df["FIELD_ID"]]
            # if vla_otf:
            #     par_df["SOURCE_NAME"] = np.array(source_tb.getcol("NAME"))[
            #         par_df["SOURCE_ID"]
            #     ]

    # Get intents and subscan numbers if available from state table.
    if table_exists(os.path.join(in_file, "STATE")):
        state_tb = tables.table(
            os.path.join(in_file, "STATE"),
            readonly=True,
            lockoptions={"option": "usernoread"},
            ack=False,
        )
        if state_tb.nrows() != 0:
            # print('state_tb',state_tb.nrows(),state_tb)
            par_df["OBS_MODE"] = np.array(state_tb.getcol("OBS_MODE"))[
                par_df["STATE_ID"]
            ]
            par_df["SUB_SCAN_NUMBER"] = state_tb.getcol("SUB_SCAN")[par_df["STATE_ID"]]
        else:
            par_df.drop(["STATE_ID"], axis=1)

    # Check if all partition scheme criteria are present in the partition table.
    partition_scheme_updated = []
    partition_criteria = {}
    for par in partition_scheme:
        if par in par_df.columns:
            partition_criteria[par] = par_df[par].unique()
            partition_scheme_updated.append(par)
    logger.info(f"Partition scheme that will be used: {partition_scheme_updated}")

    # Make all possible combinations of the partition criteria.
    enumerated_partitions = enumerated_product(*list(partition_criteria.values()))

    # print('par_df',par_df)

    # Create a list of dictionaries with the partition information. This will be used to query the MSv2 main table.
    partitions = []
    partition_axis_names = [
        "DATA_DESC_ID",
        "OBSERVATION_ID",
        "FIELD_ID",
        "SCAN_NUMBER",
        "STATE_ID",
        "SOURCE_ID",
        "OBS_MODE",
        "SUB_SCAN_NUMBER",
    ]
    if "ANTENNA1" in partition_scheme:
        partition_axis_names.append("ANTENNA1")

    for idx, pair in enumerated_partitions:
        query = ""
        for i, par in enumerate(partition_scheme_updated):
            if isinstance(pair[i], str):
                query = query + f'{par} == "{pair[i]}" and '
            else:
                query = query + f"{par} == {pair[i]} and "
        query = query[:-4]  # remove last and
        sub_par_df = par_df.query(query).drop_duplicates()

        if sub_par_df.shape[0] != 0:
            partition_info = {}

            # FIELD_NAME	SOURCE_NAME
            for col_name in partition_axis_names:
                if col_name in sub_par_df.columns:
                    partition_info[col_name] = sub_par_df[col_name].unique()
                else:
                    partition_info[col_name] = [None]

            partitions.append(partition_info)

    return partitions



"""
Processing Set and Measurement Set v4 API. Includes functions and classes to open, load,
convert, and retrieve information from Processing Set and Measurement Sets nodes of the
Processing Set DataTree
"""

import toolviper.utils.logger as _logger

from .processing_set_xdt import ProcessingSetXdt
from .open_processing_set import open_processing_set
from .load_processing_set import load_processing_set
from .measurement_set_xdt import MeasurementSetXdt
from .schema import SpectrumXds, VisibilityXds

__all__ = [
    "ProcessingSetXdt",
    "MeasurementSetXdt",
    "open_processing_set",
    "load_processing_set",
    "SpectrumXds",
    "VisibilityXds",
]

try:
    from .convert_msv2_to_processing_set import (
        convert_msv2_to_processing_set,
        estimate_conversion_memory_and_cores,
    )
except ModuleNotFoundError as exc:
    _logger.warning(
        "Could not import the function to convert from MSv2 to MSv4. "
        f"That functionality will not be available. Details: {exc}"
    )
else:
    __all__.extend(
        ["convert_msv2_to_processing_set", "estimate_conversion_memory_and_cores"]
    )



import copy
import dask.array as da

# import toolviper.utils.logger as logger
import numpy as np
import os
import xarray as xr
import s3fs
from .common import _np_types, _top_level_sub_xds
from ..common import _coords_to_numpy, _dask_arrayize_dv, _numpy_arrayize_dv
from xradio._utils.zarr.common import _get_file_system_and_items


def _read_zarr(
    zarr_store: str, id_dict: dict, selection: dict = {}
) -> (xr.Dataset, bool):
    # supported key/values in id_dict are:
    # "dv"
    #    what data variables should be returned as.
    #    "numpy": numpy arrays
    #    "dask": dask arrays
    # "coords"
    #    what coords should be returned as
    #    "numpy": numpy arrays
    # it's easiest just to copy the object rather than figuring out
    # if it should be copied at some point or not
    # xds = xr.open_zarr(zarr_store).isel(selection).copy(deep=True)
    xds = xr.open_zarr(zarr_store).isel(selection)
    do_dask = False
    do_numpy = False
    do_np_coords = False
    if "dv" in id_dict:
        dv = id_dict["dv"]
        if dv in ["dask", "numpy"]:
            do_dask = dv == "dask"
            do_numpy = not do_dask
        else:
            raise ValueError(
                f"Unsupported value {id_dict[dv]} for id_dict[dv]. "
                "Supported values are 'dask' and 'numpy'"
            )
    if "coords" in id_dict:
        c = id_dict["coords"]
        if c == "numpy":
            do_np_coords = True
        else:
            raise ValueError(
                f"Unexpected value {c} for id_dict[coords]. "
                "The supported value is 'numpy'"
            )
    # do not pass selection, because that is only for the top level data vars
    xds = _decode(xds, zarr_store, id_dict)
    if do_np_coords:
        xds = _coords_to_numpy(xds)
    if do_dask:
        xds = _dask_arrayize_dv(xds)
    elif do_numpy:
        xds = _numpy_arrayize_dv(xds)
    return xds


def _decode(xds: xr.Dataset, zarr_store: str, id_dict: dict) -> xr.Dataset:
    xds.attrs = _decode_dict(xds.attrs, "")
    _decode_sub_xdses(xds, zarr_store, id_dict)
    return xds


def _decode_dict(my_dict: dict, top_key: str) -> dict:
    # Decodes numpy arrays
    my_dict
    for k, v in my_dict.items():
        if isinstance(v, dict):
            if (
                "_type" in v
                and v["_type"] == "numpy.ndarray"
                and "_value" in v
                and "_dtype" in v
            ):
                my_dict[k] = np.array(v["_value"], dtype=_np_types[v["_dtype"]])
            else:
                z = os.sep.join([top_key, k]) if top_key else k
                my_dict[k] = _decode_dict(v, z)
    return my_dict


def _decode_sub_xdses(xarrayObj, top_dir: str, id_dict: dict) -> None:
    # FIXME this also needs to support S3
    # determine immediate subdirs of zarr_store
    entries = os.scandir(top_dir)
    for d in entries:
        path = os.path.join(top_dir, d.name)
        if os.path.isdir(path):
            if d.name.startswith(_top_level_sub_xds):
                ky = d.name[len(_top_level_sub_xds) :]
                xarrayObj.attrs[ky] = _read_zarr(path, id_dict)
                # TODO if attrs that are xdses have attrs that are xdses ...
            else:
                # descend into the directory
                _decode_sub_xdses(xarrayObj[d.name], path, id_dict)


"""
def _decode_sub_xdses(zarr_store: str, id_dict: dict) -> dict:
    sub_xdses = {}
    fs, store_contents = _get_file_system_and_items(zarr_store)
    if isinstance(fs, s3fs.core.S3FileSystem):
        # could we just use the items as returned from the helper function..?
        store_tree = fs.walk(zarr_store, topdown=True)
        # Q: what is prepend_s3 used for? In this version it is defined but not used.
        prepend_s3 = "s3://"
    else:
        store_tree = os.walk(zarr_store, topdown=True)
        prepend_s3 = ""
    for root, dirs, files in store_tree:
        relpath = os.path.relpath(root, zarr_store)
        print("rpath", relpath)
        for d in dirs:
            if d.startswith(_top_level_sub_xds):
                xds = _read_zarr(os.sep.join([root, d]), id_dict)
                # for k, v in xds.data_vars.items():
                #    xds = xds.drop_vars([k]).assign({k: v.compute()})
                ky = d[len(_top_level_sub_xds) :]
                sub_xdses[ky] = xds
    print(f"Sub xdses: {sub_xdses.keys()}")
    print("return")
    return sub_xdses
"""



import toolviper.utils.logger as logger
import os
import time
from typing import Optional, Tuple, Union

import numpy as np
import xarray as xr
from numpy.typing import ArrayLike

from xradio._utils.coord_math import convert_to_si_units
from xradio._utils.dict_helpers import (
    make_time_measure_attrs,
    make_spectral_coord_measure_attrs,
)
from xradio._utils.schema import (
    column_description_casacore_to_msv4_measure,
    convert_generic_xds_to_xradio_schema,
)
from xradio.measurement_set._utils._utils.interpolate import (
    interpolate_to_time,
)
from xradio.measurement_set._utils._msv2.subtables import subt_rename_ids
from xradio.measurement_set._utils._msv2._tables.read import (
    load_generic_table,
    make_taql_where_between_min_max,
    table_exists,
    table_has_column,
)


standard_time_coord_attrs = make_time_measure_attrs(time_format="unix")


def rename_and_interpolate_to_time(
    xds: xr.Dataset,
    time_initial_name: str,
    interp_time: Union[xr.DataArray, None],
    message_prefix: str,
) -> xr.Dataset:
    """
    This function interpolates the time dimension and renames it:

    - interpolates a time_* dimension to values given in interp_time (presumably the time
    axis of the main xds)
    - rename/replace that time_* dimension to "time", where time_* is a (sub)xds specific
    time axis
    (for example "time_pointing", "time_ephemeris", "time_syscal", "time_phase_cal").

    If interp_time is None this will simply return the input xds without modificaitons.
    Uses interpolate_to_time() for interpolation.
    ...

    Parameters:
    ----------
    xds : xr.Dataset
        Xarray dataset to interpolate (presumably a pointing_xds or an xds of
        ephemeris variables)
    time_initial_name: str = None
        Name of time to be renamed+interpolated. Expected an existing time_* coordinate in the
        dataset
    interp_time:
        Time axis to interpolate the dataset to (usually main MSv4 time)
    message_prefix:
        A prefix for info/debug/etc. messages about the specific xds being interpolated/
        time-renamed

    Returns:
    -------
    renamed_interpolated_xds : xr.Dataset
        xarray dataset with time axis renamed to "time" (from time_name, for example
        "time_ephemeris") and interpolated to interp_time.
    """
    if interp_time is None:
        return xds

    interpolated_xds = interpolate_to_time(
        xds,
        interp_time,
        message_prefix,
        time_name=time_initial_name,
    )

    # rename the time_* axis to time.
    time_coord = {"time": (time_initial_name, interp_time.data)}
    renamed_time_xds = interpolated_xds.assign_coords(time_coord)
    renamed_time_xds.coords["time"].attrs.update(standard_time_coord_attrs)
    renamed_time_xds = renamed_time_xds.swap_dims({time_initial_name: "time"})
    if time_initial_name != "time":
        renamed_time_xds = renamed_time_xds.drop_vars(time_initial_name)

    return renamed_time_xds


def make_taql_where_weather(
    in_file: str, ant_xds_station_name_ids: xr.DataArray
) -> str:
    """
    The use of taql_where with WEATHER is complicated because of ALMA and its
    (ab)use of NS_WX_STATION_ID vs. ANTENNA_ID (all -1) (see read.py). We
    cannot simply use a 'WHERE ANTENNA_ID=...'. This function produces a TaQL
    where string that uses ANTENNA_ID/NS_WX_STATION_ID depending on waht is
    found in the WEATHER subtable.

    Parameters
    ----------
    in_file : str
        Input MS name.
    ant_xds_station_name_ids : xr.DataArray
        station name data array from antenna_xds, with name/id information

    Returns
    -------
    taql_where: str
        WHERE substring safe to use in taql_where when loading WEATHER subtable
    """
    weather_path = os.path.join(in_file, "WEATHER")
    if table_exists(weather_path) and table_has_column(
        weather_path, "NS_WX_STATION_ID"
    ):
        taql_where = f"WHERE (NS_WX_STATION_ID IN [{','.join(map(str, ant_xds_station_name_ids.antenna_id.values))}])"
    else:
        taql_where = f"WHERE (ANTENNA_ID IN [{','.join(map(str, ant_xds_station_name_ids.antenna_id.values))}])"

    return taql_where


def prepare_generic_weather_xds_and_station_name(
    generic_weather_xds: xr.Dataset,
    in_file: str,
    ant_position_with_ids: xr.DataArray,
    has_asdm_station_position: bool,
) -> tuple[xr.Dataset, np.ndarray]:
    """
    A generic_weather_xds loaded with load_generic_table() might still need to be reloaded
    with an additional WHERE condition to constrain the indices of antennas. But this depends on whether
    ASDM/importasdm extension columns are present or not.

    This also prepares the station_name values:
    - if has_asdm_station_ids:
       - tries to find from ASDM_STATION the station names,
       - otherwise, takes ids (antenna_ids in generic_weather were actually the ASDM_STATION_IDs
    - else: get the values from antenna_xds (the stations present)


    Parameters
    ----------
    generic_weather_xds : xr.Dataset
        generic dataset read from an MSv2 WEATHER subtable
    in_file : str
        Input MS name.
    ant_position_with_ids : xr.DataArray
        antenna_position data var from the antenna_xds (expected to still include the initial ANTENNA_ID
        coordinate as well as other coordinates from the antenna_xds)
    has_asdm_station_position : bool
        wHether this generic weatehr_xds should be treated as including the nonstandard extensions
        NS_WX_STATION_ID and NS_WX_STATION_POSITION as created by CASA/importasdm (ALMA and VLA).

    Returns
    -------
    (generic_weather_xds, station_name): tuple[[xarray.Dataset, numpy.ndarray]
        Weather Xarray Dataset prepared for generic conversion to MSv4, values for the station_name coordinate
    """

    if has_asdm_station_position:
        asdm_station_path = os.path.join(in_file, "ASDM_STATION")
        if table_exists(asdm_station_path):
            asdm_station_xds = load_generic_table(in_file, "ASDM_STATION")
            station_name = asdm_station_xds.name.values[
                generic_weather_xds["ANTENNA_ID"].values
            ]
        else:
            # if no info from ASDM_STATION, use the indices from antenna_id which was actually the NS_WX_STATION_ID
            len_antenna_id = generic_weather_xds.sizes["ANTENNA_ID"]
            station_name = list(
                map(
                    lambda x, y: x + "_" + y,
                    ["Station"] * len_antenna_id,
                    generic_weather_xds["ANTENNA_ID"].values.astype(str),
                )
            )

    else:
        taql_where = make_taql_where_weather(in_file, ant_position_with_ids)
        generic_weather_xds = load_generic_table(
            in_file,
            "WEATHER",
            timecols=["TIME"],
            rename_ids=subt_rename_ids["WEATHER"],
            taql_where=taql_where,
        )

        if not generic_weather_xds.data_vars:
            # for example when the weather subtable only has info for antennas/stations
            # not present in the MSv4 (no overlap between antennas loaded in ant_xds and weather)
            return None, None

        stations_present = ant_position_with_ids.sel(
            antenna_id=generic_weather_xds["ANTENNA_ID"]
        ).station_name
        station_name = stations_present.values

    return generic_weather_xds, station_name


def finalize_station_position(
    weather_xds: xr.Dataset, ant_position_with_ids, has_asdm_station_position: bool
) -> xr.Dataset:
    """
    For a STATION_POSITION data var being added to a weather_xds, make sure coordinates and dimensions
    are conforming to the schema.

    Parameters
    ----------
    weather_xds : xr.Dataset
        weather_xds where we still need to ensure the right coordinates and attributes
    ant_position_with_ids : xr.DataArray
        antenna_position data var from the antenna_xds (expected to still include the initial ANTENNA_ID
        coordinate as well as other coordinates from the antenna_xds)
    has_asdm_station_position : bool
        Whether this generic weatehr_xds should be treated as including the nonstandard extensions
        NS_WX_STATION_ID and NS_WX_STATION_POSITION as created by CASA/importasdm (ALMA and VLA).

    Returns
    -------
    weather_xds: xarray.Dataset
        Weather Xarray Dataset with all coordinates and attributes in STATION_POSITION
    """
    if has_asdm_station_position:
        # STATION_POSITION has been created but needs prooper dimensions and attrs
        # Drop the time dim
        weather_xds["STATION_POSITION"] = weather_xds["STATION_POSITION"].sel(
            time_weather=0, drop=True, method="nearest"
        )
        # borrow location frame attributes from antenna position
        weather_xds["STATION_POSITION"].attrs = ant_position_with_ids.attrs
    else:
        # borrow from ant_position_with_ids but without carrying over other coords
        weather_xds = weather_xds.assign(
            {
                "STATION_POSITION": (
                    ["station_name", "cartesian_pos_label"],
                    ant_position_with_ids.values,
                    ant_position_with_ids.attrs,
                )
            }
        )

    return weather_xds


def create_weather_xds(in_file: str, ant_position_with_ids: xr.DataArray):
    """
    Creates a Weather Xarray Dataset from a MS v2 WEATHER table.

    Parameters
    ----------
    in_file : str
        Input MS name.
    ant_position_with_ids : xr.DataArray
        antenna_position data var from the antenna_xds (expected to still including the initial ANTENNA_ID coordinate
        as wellas other coordinates from the antenna_xds)

    Returns
    -------
    xr.Dataset
        Weather Xarray Dataset.
    """

    try:
        generic_weather_xds = load_generic_table(
            in_file,
            "WEATHER",
            timecols=["TIME"],
            rename_ids=subt_rename_ids["WEATHER"],
        )
    except ValueError as _exc:
        return None

    has_asdm_station_position = (
        "NS_WX_STATION_POSITION" in generic_weather_xds.data_vars
    )
    generic_weather_xds, station_name = prepare_generic_weather_xds_and_station_name(
        generic_weather_xds, in_file, ant_position_with_ids, has_asdm_station_position
    )
    if not generic_weather_xds:
        return None

    weather_xds = xr.Dataset(attrs={"type": "weather"})
    coords = {
        "station_name": station_name,
        "cartesian_pos_label": ["x", "y", "z"],
    }
    weather_xds = weather_xds.assign_coords(coords)

    dims_station_time = ["station_name", "time_weather"]
    dims_station_time_position = dims_station_time + ["cartesian_pos_label"]
    to_new_data_variables = {
        "H2O": ["H2O", dims_station_time],
        "IONOS_ELECTRON": ["IONOS_ELECTRON", dims_station_time],
        "PRESSURE": ["PRESSURE", dims_station_time],
        "REL_HUMIDITY": ["REL_HUMIDITY", dims_station_time],
        "TEMPERATURE": ["TEMPERATURE", dims_station_time],
        "DEW_POINT": ["DEW_POINT", dims_station_time],
        "WIND_DIRECTION": ["WIND_DIRECTION", dims_station_time],
        "WIND_SPEED": ["WIND_SPEED", dims_station_time],
    }
    if has_asdm_station_position:
        to_new_data_variables.update(
            {
                "NS_WX_STATION_POSITION": [
                    "STATION_POSITION",
                    dims_station_time_position,
                ],
            }
        )

    to_new_coords = {
        "TIME": ["time_weather", ["time_weather"]],
    }

    weather_xds = convert_generic_xds_to_xradio_schema(
        generic_weather_xds, weather_xds, to_new_data_variables, to_new_coords
    )
    weather_xds = finalize_station_position(
        weather_xds, ant_position_with_ids, has_asdm_station_position
    )

    # TODO: option to interpolate to main time

    # PRESSURE: hPa in MSv2 specs and some MSs => Pa
    weather_xds = convert_to_si_units(weather_xds)

    # correct expected types (for example "IONOS_ELECTRON", "PRESSURE" can be float32)
    for data_var in weather_xds:
        if weather_xds.data_vars[data_var].dtype != np.float64:
            weather_xds[data_var] = weather_xds[data_var].astype(np.float64)
    if "time_weather" in weather_xds.coords:
        weather_xds.coords["time_weather"].attrs["type"] = "time_weather"

    return weather_xds


def correct_generic_pointing_xds(
    generic_pointing_xds: xr.Dataset, to_new_data_variables: dict[str, list]
) -> xr.Dataset:
    """
    Takes a (generic) pointing_xds as read from a POINTING subtable of an MSv2
    and tries to correct several deviations from the MSv2 specs seen in
    common test data.
    The problems fixed here include wrong dimensions:

    - for example transposed dimensions with respect to the MSv2 specs (output
    from CASA simulator),
    - missing/additional unexpected dimensions when some of the columns are
    empty (in the sense of "empty casacore cells").

    This function modifies the data arrays of the data vars affected by such
    issues.

    Parameters
    ----------
    generic_pointing_xds: xr.Dataset
        The generic pointing dataset (loaded from MSv2) to be fixed
    to_new_data_variables: dict
        The dict used for convert_generic_xds_to_xradio_schema, which gives all
        the data variables relevant for the final MSv4 dataset.

    Returns:
    --------
    xr.Dataset
        Corrected dataset with dimensions conforming to MSv2 specs.
    """

    correct_pointing_xds = generic_pointing_xds.copy()

    for data_var_name in generic_pointing_xds:
        if data_var_name in to_new_data_variables:
            # Corrects dim sizes of "empty cell" variables, such as empty DIRECTION, TARGET, etc.
            if (
                "dim_2" in generic_pointing_xds.sizes
                and generic_pointing_xds.sizes["dim_2"] == 0
            ):
                # When all direction variables are "empty"
                data_var_data = xr.DataArray(
                    [[[[np.nan, np.nan]]]],
                    dims=generic_pointing_xds.dims,
                ).isel(n_polynomial=0, drop=True)
                correct_pointing_xds[data_var_name].data = data_var_data

            elif (
                "dir" in generic_pointing_xds.sizes
                and generic_pointing_xds.sizes["dir"] == 0
            ):
                # When some direction variables are "empty" but some are populated properly
                if "dim_2" in generic_pointing_xds[data_var_name].sizes:
                    data_var_data = xr.DataArray(
                        generic_pointing_xds[data_var_name].values,
                        dims=generic_pointing_xds[data_var_name].dims,
                    )
                else:
                    shape = tuple(
                        generic_pointing_xds.sizes[dim]
                        for dim in ["TIME", "ANTENNA_ID"]
                    ) + (2,)
                    data_var_data = xr.DataArray(
                        np.full(shape, np.nan),
                        dims=generic_pointing_xds[data_var_name].dims,
                    )
                correct_pointing_xds[data_var_name].data = data_var_data

    return correct_pointing_xds


def create_pointing_xds(
    in_file: str,
    ant_xds_name_ids: xr.DataArray,
    time_min_max: Union[Tuple[np.float64, np.float64], None],
    interp_time: Union[xr.DataArray, None] = None,
) -> xr.Dataset:
    """
    Creates a Pointing Xarray Dataset from an MS v2 POINTING (sub)table.

    WIP: details of a few direction variables (and possibly moving some to attributes) to be
    settled (see MSv4 spreadsheet).

    Parameters
    ----------
    in_file : str
        Input MS name.
    ant_xds_name_ids : xr.DataArray
        antenna_name data array from antenna_xds, with name/id information
    time_min_max : tuple
        min / max times values to constrain loading (from the TIME column)
    interp_time : Union[xr.DataArray, None] (Default value = None)
        interpolate time to this (presumably main dataset time)

    Returns
    -------
    xr.Dataset
         Pointing Xarray dataset
    """
    start = time.time()

    taql_time_range = make_taql_where_between_min_max(
        time_min_max, in_file, "POINTING", "TIME"
    )

    if taql_time_range is None:
        taql_where = f"WHERE (ANTENNA_ID IN [{','.join(map(str, ant_xds_name_ids.antenna_id.values))}])"
    else:
        taql_where = (
            taql_time_range
            + f" AND (ANTENNA_ID IN [{','.join(map(str, ant_xds_name_ids.antenna_id.values))}])"
        )
    # Read POINTING table into a Xarray Dataset.
    generic_pointing_xds = load_generic_table(
        in_file,
        "POINTING",
        timecols=["TIME"],
        rename_ids=subt_rename_ids["POINTING"],
        taql_where=taql_where,
    )

    if not generic_pointing_xds.data_vars:
        # apparently empty MS/POINTING table => produce empty xds
        return xr.Dataset()

    # Checking a simple way of using only the one single coefficient of the polynomials
    if "n_polynomial" in generic_pointing_xds.sizes:
        size = generic_pointing_xds.sizes["n_polynomial"]
        if size == 1:
            generic_pointing_xds = generic_pointing_xds.sel({"n_polynomial": 0})
        elif size == 0:
            generic_pointing_xds = generic_pointing_xds.drop_dims("n_polynomial")

    time_ant_dims = ["time_pointing", "antenna_name"]
    time_ant_dir_dims = time_ant_dims + ["local_sky_dir_label"]
    to_new_data_variables = {
        "DIRECTION": ["POINTING_BEAM", time_ant_dir_dims],
        "ENCODER": ["POINTING_DISH_MEASURED", time_ant_dir_dims],
        "OVER_THE_TOP": ["POINTING_OVER_THE_TOP", time_ant_dims],
    }

    to_new_coords = {
        "TIME": ["time_pointing", ["time_pointing"]],
        "dim_2": ["local_sky_dir_label", ["local_sky_dir_label"]],
    }

    generic_pointing_xds = correct_generic_pointing_xds(
        generic_pointing_xds, to_new_data_variables
    )
    pointing_xds = xr.Dataset(attrs={"type": "pointing"})
    coords = {
        "antenna_name": ant_xds_name_ids.sel(
            antenna_id=generic_pointing_xds["ANTENNA_ID"]
        ).data,
        "local_sky_dir_label": ["az", "alt"],
    }
    pointing_xds = pointing_xds.assign_coords(coords)
    pointing_xds = convert_generic_xds_to_xradio_schema(
        generic_pointing_xds, pointing_xds, to_new_data_variables, to_new_coords
    )

    pointing_xds = rename_and_interpolate_to_time(
        pointing_xds, "time_pointing", interp_time, "pointing_xds"
    )

    logger.debug(f"create_pointing_xds() execution time {time.time() - start:0.2f} s")

    return pointing_xds


def prepare_generic_sys_cal_xds(generic_sys_cal_xds: xr.Dataset) -> xr.Dataset:
    """
    A generic_sys_cal_xds loaded with load_generic_table() cannot be easily
    used in convert_generic_xds_to_xradio_schema() to produce an MSv4
    sys_cal_xds dataset, as their structure differs in dimensions and order
    of dimensions.
    This function performs various prepareation steps, such as:

    - filter out dimensions not neeed for an individual MSv4 (SPW, FEED),
    - drop variables loaded from columns with all items set to empty array,
    - transpose the dimensions frequency,receptor
    - fix dimension names (and order) when needed.

    Parameters
    ----------
    generic_sys_cal_xds : xr.Dataset
        generic dataset read from an MSv2 SYSCAL subtable

    Returns
    -------
    generic_sys_cal_xds: xr.Dataset
        System calibration Xarray Dataset prepared for generic conversion
        to MSv4.
    """

    # drop SPW and feed dims
    generic_sys_cal_xds = generic_sys_cal_xds.isel(SPECTRAL_WINDOW_ID=0, drop=True)
    generic_sys_cal_xds = generic_sys_cal_xds.isel(FEED_ID=0, drop=True)

    # Often some of the T*_SPECTRUM are present but all the cells are populated
    # with empty arrays
    empty_arrays_vars = []
    for data_var in generic_sys_cal_xds.data_vars:
        if generic_sys_cal_xds[data_var].size == 0:
            empty_arrays_vars.append(data_var)
    if empty_arrays_vars:
        generic_sys_cal_xds = generic_sys_cal_xds.drop_vars(empty_arrays_vars)

    # Re-arrange receptor and frequency dims depending on input structure
    if (
        "receptor" in generic_sys_cal_xds.sizes
        and "frequency" in generic_sys_cal_xds.sizes
    ):
        # dim_3 can be created for example when the T*_SPECTRUM have varying # channels!
        # more generaly, could transpose with ... to avoid errors with additional spurious dimensions
        if "dim_3" in generic_sys_cal_xds.dims:
            generic_sys_cal_xds = generic_sys_cal_xds.drop_dims("dim_3")
        # From MSv2 tables we get (...,frequency, receptor)
        #  -> transpose to (...,receptor,frequency) ready for MSv4 sys_cal_xds
        generic_sys_cal_xds = generic_sys_cal_xds.transpose(
            "ANTENNA_ID", "TIME", "receptor", "frequency"
        )
    elif (
        "frequency" in generic_sys_cal_xds.sizes
        and "dim_3" not in generic_sys_cal_xds.sizes
    ):
        # because order is (...,frequency,receptor), when frequency is missing
        # receptor can get wrongly labeled as frequency
        generic_sys_cal_xds = generic_sys_cal_xds.rename_dims({"frequency": "receptor"})
    elif (
        "frequency" not in generic_sys_cal_xds.sizes
        and "receptor" in generic_sys_cal_xds.sizes
        and "dim_3" in generic_sys_cal_xds.sizes
    ):
        # different *_SPECTRUM array sizes + some empty arrays can create an additional spurious
        # generic dimension, which should have been "receptor"
        generic_sys_cal_xds = generic_sys_cal_xds.rename_dims({"receptor": "frequency"})
        generic_sys_cal_xds = generic_sys_cal_xds.rename_dims({"dim_3": "receptor"})
        generic_sys_cal_xds = generic_sys_cal_xds.transpose(
            "ANTENNA_ID", "TIME", "receptor", "frequency"
        )
    else:
        raise RuntimeError(
            "Cannot understand the arrangement of dimensions of {generic_sys_cal_xds=}"
        )

    return generic_sys_cal_xds


def create_system_calibration_xds(
    in_file: str,
    spectral_window_id: int,
    main_xds_frequency: xr.DataArray,
    ant_xds: xr.DataArray,
    sys_cal_interp_time: Union[xr.DataArray, None] = None,
):
    """
    Creates a system calibration Xarray Dataset from a MSv2 SYSCAL table.

    Parameters
    ----------
    in_file: str
        Input MS name.
    spectral_window_id: int
        SPW ID from MSv2
    main_xds_frequency: xr.DataArray
        frequency array of the main xds (MSv4), containing among other things
        the measures metadata
    ant_xds : xr.Dataset
        The antenna_xds that has information such as names, stations, etc., for coordinates
    sys_cal_interp_time: Union[xr.DataArray, None] = None,
        Time axis to interpolate the data vars to (usually main MSv4 time)

    Returns
    -------
    sys_cal_xds: xr.Dataset
        System calibration Xarray Dataset.
    """

    try:
        generic_sys_cal_xds = load_generic_table(
            in_file,
            "SYSCAL",
            timecols=["TIME"],
            rename_ids=subt_rename_ids["SYSCAL"],
            taql_where=(
                f" where (SPECTRAL_WINDOW_ID = {spectral_window_id})"
                f" AND (ANTENNA_ID IN [{','.join(map(str, ant_xds.antenna_id.values))}])"
            ),
        )
    except ValueError as _exc:
        return None

    if not generic_sys_cal_xds.data_vars:
        # even though SYSCAL is an optional subtable, some write it empty
        return None

    generic_sys_cal_xds = prepare_generic_sys_cal_xds(generic_sys_cal_xds)

    mandatory_dimensions = ["antenna_name", "time_system_cal", "receptor_label"]
    if "frequency" not in generic_sys_cal_xds.sizes:
        dims_all = mandatory_dimensions
    else:
        dims_all = mandatory_dimensions + ["frequency_system_cal"]

    to_new_data_variables = {
        "PHASE_DIFF": ["PHASE_DIFFERENCE", ["antenna_name", "time_system_cal"]],
        "TCAL": ["TCAL", dims_all],
        "TCAL_SPECTRUM": ["TCAL", dims_all],
        "TRX": ["TRX", dims_all],
        "TRX_SPECTRUM": ["TRX", dims_all],
        "TSKY": ["TSKY", dims_all],
        "TSKY_SPECTRUM": ["TSKY", dims_all],
        "TSYS": ["TSYS", dims_all],
        "TSYS_SPECTRUM": ["TSYS", dims_all],
        "TANT": ["TANT", dims_all],
        "TANT_SPECTRUM": ["TANT", dims_all],
        "TAN_TSYS": ["TAN_TSYS", dims_all],
        "TANT_SYS_SPECTRUM": ["TANT_TSYS", dims_all],
    }

    to_new_coords = {
        "TIME": ["time_system_cal", ["time_system_cal"]],
        "receptor": ["receptor_label", ["receptor_label"]],
        "frequency": ["frequency_system_cal", ["frequency_system_cal"]],
    }

    sys_cal_xds = xr.Dataset(attrs={"type": "system_calibration"})
    ant_borrowed_coords = {
        "antenna_name": ant_xds.coords["antenna_name"],
        "receptor_label": ant_xds.coords["receptor_label"],
        "polarization_type": ant_xds.coords["polarization_type"],
    }
    sys_cal_xds = sys_cal_xds.assign_coords(ant_borrowed_coords)
    sys_cal_xds = convert_generic_xds_to_xradio_schema(
        generic_sys_cal_xds, sys_cal_xds, to_new_data_variables, to_new_coords
    )

    # Add frequency coord and its measures data, if present
    if "frequency_system_cal" in dims_all:
        frequency_coord = {
            "frequency_system_cal": generic_sys_cal_xds.coords["frequency"].data
        }
        sys_cal_xds = sys_cal_xds.assign_coords(frequency_coord)
        frequency_measure = make_spectral_coord_measure_attrs(
            main_xds_frequency.attrs["units"], main_xds_frequency.attrs["observer"]
        )
        sys_cal_xds.coords["frequency_system_cal"].attrs.update(frequency_measure)

    sys_cal_xds = rename_and_interpolate_to_time(
        sys_cal_xds, "time_system_cal", sys_cal_interp_time, "system_calibration_xds"
    )

    # correct expected types
    for data_var in sys_cal_xds:
        if sys_cal_xds.data_vars[data_var].dtype != np.float64:
            sys_cal_xds[data_var] = sys_cal_xds[data_var].astype(np.float64)

    return sys_cal_xds


def create_phased_array_xds(
    in_file: str,
    antenna_names: list[str],
    receptor_label: list[str],
    polarization_type: ArrayLike,
) -> Optional[xr.Dataset]:
    """
    Create an Xarray Dataset containing phased array information.

    Parameters
    ----------
    in_file : str
        Path to the input MSv2.
    antenna_names: DataArray or Sequence[str]
        Content of the antenna_name coordinate of the antenna_xds.
    receptor_label: DataArray or Sequence[str]
        Content of the receptor_label coordinate of the antenna_xds. Used to
        label the corresponding axis of ELEMENT_FLAG.
    polarization_type: DataArray or ArrayLike
        Contents of the polarization_type coordinate of the antenna_xds.
        Array-like of shape (num_antennas, 2) containing the polarization
        hands for each antenna.

    Returns
    ----------
        xr.Dataset or None: If the input MS contains a PHASED_ARRAY table,
           returns the Xarray Dataset containing the phased array information.
           Otherwise, return None.
    """

    def extract_data(dataarray_or_sequence):
        if hasattr(dataarray_or_sequence, "data"):
            return dataarray_or_sequence.data.tolist()
        return dataarray_or_sequence

    antenna_names = extract_data(antenna_names)
    receptor_label = extract_data(receptor_label)
    polarization_type = extract_data(polarization_type)

    # NOTE: We cannot use the dimension renaming option of `load_generic_table`
    # here, because it leads to a dimension name collision. This is caused by
    # the presence of two dimensions of size 3 in multiple arrays.
    # Instead, we do the renaming manually below.
    try:
        raw_xds = load_generic_table(
            in_file,
            "PHASED_ARRAY",
            # Some MSes carry COORDINATE_SYSTEM as a copy of COORDINATE_AXES
            # due to a past ambiguity on the PHASED_ARRAY schema
            ignore=["COORDINATE_SYSTEM", "ANTENNA_ID"],
        )
    except ValueError:
        return None

    # Defend against empty PHASED_ARRAY table.
    # The test MS "AA2-Mid-sim_00000.ms" has that problem.
    required_keys = {"COORDINATE_AXES", "ELEMENT_OFFSET", "ELEMENT_FLAG"}
    if not all(k in raw_xds for k in required_keys):
        return None

    def msv4_measure(raw_name: str) -> dict:
        coldesc = raw_xds.attrs["other"]["msv2"]["ctds_attrs"]["column_descriptions"]
        return column_description_casacore_to_msv4_measure(coldesc[raw_name])

    def make_data_variable(raw_name: str, dim_names: list[str]) -> xr.DataArray:
        da = raw_xds[raw_name]
        da = xr.DataArray(da.data, dims=tuple(dim_names))
        return da.assign_attrs(msv4_measure(raw_name))

    raw_datavar_names_and_dims = [
        (
            "COORDINATE_AXES",
            ("antenna_name", "cartesian_pos_label_local", "cartesian_pos_label"),
        ),
        ("ELEMENT_OFFSET", ("antenna_name", "cartesian_pos_label_local", "element_id")),
        ("ELEMENT_FLAG", ("antenna_name", "receptor_label", "element_id")),
    ]

    data_vars = {
        name: make_data_variable(name, dims)
        for name, dims in raw_datavar_names_and_dims
    }
    data_vars["COORDINATE_AXES"].attrs = {
        "type": "rotation_matrix",
        "units": "dimensionless",
    }
    # Remove the "frame" attribute if it exists, because ELEMENT_OFFSET is
    # defined in a station-local frame for which no standard name exists
    data_vars["ELEMENT_OFFSET"].attrs.pop("frame", None)
    data_vars["ELEMENT_OFFSET"].attrs.update(
        {
            "coordinate_system": "topocentric",
            "origin": "ANTENNA_POSITION",
        }
    )

    num_elements = data_vars["ELEMENT_OFFSET"].sizes["element_id"]

    data_vars = {"PHASED_ARRAY_" + key: val for key, val in data_vars.items()}
    coords = {
        "antenna_name": antenna_names,
        "element_id": np.arange(num_elements),
        "receptor_label": receptor_label,
        "polarization_type": (
            ("antenna_name", "receptor_label"),
            polarization_type,
        ),
        "cartesian_pos_label": ["x", "y", "z"],
        "cartesian_pos_label_local": ["p", "q", "r"],
    }
    attrs = {"type": "phased_array"}
    return xr.Dataset(data_vars, coords, attrs)



from astropy.wcs import WCS
from collections import OrderedDict
import numpy as np
import xarray as xr
from typing import List, Union
from .common import _c, _compute_world_sph_dims, _l_m_attr_notes
from xradio._utils.coord_math import _deg_to_rad
from xradio._utils.dict_helpers import (
    make_spectral_coord_reference_dict,
    make_quantity,
    make_skycoord_dict,
    make_time_coord_attrs,
)


def _input_checks(
    phase_center: Union[list, np.ndarray],
    image_size: Union[list, np.ndarray],
    cell_size: Union[list, np.ndarray],
) -> None:
    if len(image_size) != 2:
        raise ValueError("image_size must have exactly two elements")
    if len(phase_center) != 2:
        raise ValueError("phase_center must have exactly two elements")
    if len(cell_size) != 2:
        raise ValueError("cell_size must have exactly two elements")


def _make_coords(
    chan_coords: Union[list, np.ndarray],
    time_coords: Union[list, np.ndarray],
) -> dict:
    if not isinstance(chan_coords, list) and not isinstance(chan_coords, np.ndarray):
        chan_coords = [chan_coords]
    chan_coords = np.array(chan_coords, dtype=np.float64)
    restfreq = chan_coords[len(chan_coords) // 2]
    vel_coords = (1 - chan_coords / restfreq) * _c.to("m/s").value
    if not isinstance(time_coords, list) and not isinstance(time_coords, np.ndarray):
        time_coords = [time_coords]
    time_coords = np.array(time_coords, dtype=np.float64)
    return dict(chan=chan_coords, vel=vel_coords, time=time_coords, restfreq=restfreq)


def _add_common_attrs(
    xds: xr.Dataset,
    restfreq: float,
    spectral_reference: str,
    direction_reference: str,
    phase_center: Union[List[float], np.ndarray],
    cell_size: Union[List[float], np.ndarray],
    projection: str,
) -> xr.Dataset:
    xds.time.attrs = make_time_coord_attrs(units="d", scale="utc", time_format="mjd")
    freq_vals = np.array(xds.frequency)
    xds.frequency.attrs = {
        "observer": spectral_reference.lower(),
        "reference_frequency": make_spectral_coord_reference_dict(
            value=freq_vals[len(freq_vals) // 2].item(),
            units="Hz",
            observer=spectral_reference.lower(),
        ),
        "rest_frequencies": make_quantity(restfreq, "Hz"),
        "rest_frequency": make_quantity(restfreq, "Hz"),
        "type": "spectral_coord",
        "units": "Hz",
        "wave_units": "mm",
    }
    xds.velocity.attrs = {"doppler_type": "radio", "type": "doppler", "units": "m/s"}
    reference = make_skycoord_dict(
        data=phase_center, units="rad", frame=direction_reference
    )
    reference["attrs"].update({"equinox": "j2000.0"})
    xds.attrs = {
        "data_groups": {"base": {}},
        "direction": {
            "reference": reference,
            "lonpole": make_quantity(np.pi, "rad", ["l", "m"]),
            "latpole": make_quantity(0.0, "rad", ["l", "m"]),
            "pc": [[1.0, 0.0], [0.0, 1.0]],
            "projection": projection,
            "projection_parameters": [0.0, 0.0],
        },
    }
    return xds


def _make_common_coords(
    pol_coords: Union[list, np.ndarray],
    chan_coords: Union[list, np.ndarray],
    time_coords: Union[list, np.ndarray],
) -> dict:
    some_coords = _make_coords(chan_coords, time_coords)
    return {
        "coords": {
            "time": some_coords["time"],
            "frequency": some_coords["chan"],
            "velocity": ("frequency", some_coords["vel"]),
            "polarization": pol_coords,
        },
        "restfreq": some_coords["restfreq"],
    }


def _make_lm_values(
    image_size: Union[list, np.ndarray],
    cell_size: Union[list, np.ndarray],
) -> dict:
    # l follows RA as far as increasing/decreasing, see AIPS Meme 27, change in alpha
    # definition three lines below Figure 2 and the first of the pair of equations 10.
    l = [
        (i - image_size[0] // 2) * (-1) * abs(cell_size[0])
        for i in range(image_size[0])
    ]
    m = [(i - image_size[1] // 2) * abs(cell_size[1]) for i in range(image_size[1])]
    return {"l": l, "m": m}


def _make_sky_coords(
    projection: str,
    image_size: Union[list, np.ndarray],
    cell_size: Union[list, np.ndarray],
    phase_center: Union[list, np.ndarray],
) -> dict:
    long, lat = _compute_world_sph_dims(
        projection=projection,
        shape=image_size,
        ctype=["RA", "Dec"],
        crpix=[image_size[0] // 2, image_size[1] // 2],
        crval=phase_center,
        cdelt=[-abs(cell_size[0]), abs(cell_size[1])],
        cunit=["rad", "rad"],
    )["value"]
    return {"right_ascension": (("l", "m"), long), "declination": (("l", "m"), lat)}


def _add_lm_coord_attrs(xds: xr.Dataset) -> xr.Dataset:
    attr_note = _l_m_attr_notes()
    xds.l.attrs = {
        "note": attr_note["l"],
    }
    xds.m.attrs = {
        "note": attr_note["m"],
    }


def _make_empty_sky_image(
    phase_center: Union[list, np.ndarray],
    image_size: Union[list, np.ndarray],
    cell_size: Union[list, np.ndarray],
    chan_coords: Union[list, np.ndarray],
    pol_coords: Union[list, np.ndarray],
    time_coords: Union[list, np.ndarray],
    direction_reference: str,
    projection: str,
    spectral_reference: str,
    do_sky_coords: bool,
) -> xr.Dataset:
    _input_checks(phase_center, image_size, cell_size)
    cc = _make_common_coords(pol_coords, chan_coords, time_coords)
    coords = cc["coords"]
    lm_values = _make_lm_values(image_size, cell_size)
    coords.update(lm_values)
    if do_sky_coords:
        coords.update(_make_sky_coords(projection, image_size, cell_size, phase_center))
    xds = xr.Dataset(coords=coords)
    xds = _move_beam_param_dim_coord(xds)
    _add_lm_coord_attrs(xds)
    _add_common_attrs(
        xds,
        cc["restfreq"],
        spectral_reference,
        direction_reference,
        phase_center,
        cell_size,
        projection,
    )
    return xds


def _make_uv_coords(
    xds: xr.Dataset,
    image_size: Union[list, np.ndarray],
    sky_image_cell_size: Union[list, np.ndarray],
) -> dict:
    uv_values = _make_uv_values(image_size, sky_image_cell_size)
    xds = xds.assign_coords(uv_values)
    attr = make_quantity(0.0, "lambda")
    xds.u.attrs = attr.copy()
    xds.v.attrs = attr.copy()
    return xds


def _make_uv_values(
    image_size: Union[list, np.ndarray],
    sky_image_cell_size: Union[list, np.ndarray],
) -> dict:
    im_size_wave = 1 / np.array(sky_image_cell_size)
    uv_cell_size = im_size_wave / np.array(image_size)
    u_vals = [
        (i - image_size[0] // 2) * abs(uv_cell_size[0]) for i in range(image_size[0])
    ]
    v_vals = [
        (i - image_size[1] // 2) * abs(uv_cell_size[1]) for i in range(image_size[1])
    ]
    return {"u": u_vals, "v": v_vals}


def _make_empty_aperture_image(
    phase_center: Union[list, np.ndarray],
    image_size: Union[list, np.ndarray],
    sky_image_cell_size: Union[list, np.ndarray],
    chan_coords: Union[list, np.ndarray],
    pol_coords: Union[list, np.ndarray],
    time_coords: Union[list, np.ndarray],
    direction_reference: str,
    projection: str,
    spectral_reference: str,
) -> xr.Dataset:
    _input_checks(phase_center, image_size, sky_image_cell_size)
    cc = _make_common_coords(pol_coords, chan_coords, time_coords)
    coords = cc["coords"]
    xds = xr.Dataset(coords=coords)
    xds = _make_uv_coords(xds, image_size, sky_image_cell_size)
    _add_common_attrs(
        xds,
        cc["restfreq"],
        spectral_reference,
        direction_reference,
        phase_center,
        sky_image_cell_size,
        projection,
    )
    xds = _move_beam_param_dim_coord(xds)
    return xds


def _move_beam_param_dim_coord(xds: xr.Dataset) -> xr.Dataset:
    return xds.assign_coords(beam_param=("beam_param", ["major", "minor", "pa"]))


def _make_empty_lmuv_image(
    phase_center: Union[list, np.ndarray],
    image_size: Union[list, np.ndarray],
    sky_image_cell_size: Union[list, np.ndarray],
    chan_coords: Union[list, np.ndarray],
    pol_coords: Union[list, np.ndarray],
    time_coords: Union[list, np.ndarray],
    direction_reference: str,
    projection: str,
    spectral_reference: str,
    do_sky_coords: bool,
) -> xr.Dataset:
    xds = _make_empty_sky_image(
        phase_center,
        image_size,
        sky_image_cell_size,
        chan_coords,
        pol_coords,
        time_coords,
        direction_reference,
        projection,
        spectral_reference,
        do_sky_coords,
    )
    xds = _make_uv_coords(xds, image_size, sky_image_cell_size)
    xds = _move_beam_param_dim_coord(xds)
    return xds



from typing import Generator
from contextlib import contextmanager

try:
    from casacore import tables
except ImportError:
    import xradio._utils._casacore.casacore_from_casatools as tables


@contextmanager
def open_table_ro(infile: str) -> Generator[tables.table, None, None]:
    table = tables.table(
        infile, readonly=True, lockoptions={"option": "usernoread"}, ack=False
    )
    try:
        yield table
    finally:
        table.close()


@contextmanager
def open_query(table: tables.table, query: str) -> Generator[tables.table, None, None]:

    if hasattr(tables, "taql"):
        ttq = tables.taql(query)
    else:
        ttq = table.taql(query)
    try:
        yield ttq
    finally:
        ttq.close()


class TableManager:

    def __init__(
        self,
        infile: str,
        taql_where: str = "",
    ):
        self.infile = infile
        self.taql_where = taql_where
        self.taql_query = taql_where.replace("where ", "")

    def get_table(self):
        # Performance note:
        # table.query("(DATA_DESC_ID = 0)") is slightly faster than
        # tables.taql("select * from $table (DATA_DESC_ID = 0)")
        with tables.table(
            self.infile, readonly=True, lockoptions={"option": "usernoread"}, ack=False
        ) as mtable:
            query = f"select * from $mtable {self.taql_where}"
            if hasattr(tables, "taql"):
                return tables.taql(query)
            else:
                return mtable.taql(query)



from typing import Union

import xarray as xr

import toolviper.utils.logger as logger


def interpolate_to_time(
    xds: xr.Dataset,
    interp_time: Union[xr.DataArray, None],
    message_prefix: str,
    time_name: str = "time",
) -> xr.Dataset:
    """
    Interpolate the time coordinate of the input xarray dataset to the
    a data array. This can be used for example to interpolate a pointing_xds
    to the time coord of the (main) MSv4, or similarly the ephemeris
    data variables of a field_and_source_xds.

    Uses interpolation method "linear", unless the source number of points is
    1 in which case "nearest" is used, to avoid divide-by-zero issues.

    Parameters:
    ----------
    xds : xr.Dataset
        Xarray dataset to interpolate (presumably a pointing_xds or an xds of
        ephemeris variables)
    interp_time : Union[xr.DataArray, None]
        Time axis to interpolate the dataset to (usually main MSv4 time)
    message_prefix: str
        A prefix for info/debug/etc. messages

    Returns:
    -------
    interpolated_xds : xr.Dataset
        xarray dataset with time axis interpolated to interp_time.
    """
    if interp_time is not None:
        points_before = xds[time_name].size
        if points_before > 1:
            method = "linear"
        else:
            method = "nearest"
        xds = xds.interp(
            {time_name: interp_time.data}, method=method, assume_sorted=True
        )
        # scan_name sneaks in as a coordinate of the main time axis, drop it
        if (
            "type" in xds.attrs
            and xds.attrs["type"] not in ["visibility", "spectrum", "wvr"]
            and "scan_name" in xds.coords
        ):
            xds = xds.drop_vars("scan_name")
        points_after = xds[time_name].size
        logger.debug(
            f"{message_prefix}: interpolating the time coordinate "
            f"from {points_before} to {points_after} points"
        )

    return xds



from typing import Tuple

import numpy as np
import pandas as pd

try:
    from casacore import tables
except ImportError:
    import xradio._utils._casacore.casacore_from_casatools as tables


from .table_query import open_query
from xradio._utils.list_and_array import (
    unique_1d,
    pairing_function,
    inverse_pairing_function,
)


def get_utimes_tol(mtable: tables.table, taql_where: str) -> Tuple[np.ndarray, float]:
    taql_utimes = f"select DISTINCT TIME from $mtable {taql_where}"
    with open_query(mtable, taql_utimes) as query_utimes:
        utimes = unique_1d(query_utimes.getcol("TIME", 0, -1))
        # add a tol around the time ranges returned by taql
        if len(utimes) < 2:
            tol = 1e-5
        else:
            tol = np.diff(utimes).min() / 4

    return utimes, tol


def get_baselines(tb_tool: tables.table) -> np.ndarray:
    """
    Gets the unique baselines from antenna 1 and antenna 2 ids.

    Uses a pairing function and inverse pairing function to decrease the
    computation time of finding unique values.

    Parameters
    ----------
    tb_tool : tables.table
        MeasurementSet table to get the antenna ids.

    Returns
    -------
    unique_baselines : np.ndarray
        a 2D array of unique antenna pairs
        (baselines) from the MeasurementSet table provided.
    """
    ant1, ant2 = tb_tool.getcol("ANTENNA1", 0, -1), tb_tool.getcol("ANTENNA2", 0, -1)

    baselines = np.column_stack((ant1, ant2))

    # Using pairing function to reduce the computation time of finding unique values.
    baselines_paired = pairing_function(baselines)
    unique_baselines_paired = pd.unique(baselines_paired)
    unique_baselines = inverse_pairing_function(unique_baselines_paired)

    # Sorting the unique baselines.
    unique_baselines = unique_baselines[unique_baselines[:, 1].argsort()]
    unique_baselines = unique_baselines[
        unique_baselines[:, 0].argsort(kind="mergesort")
    ]

    return unique_baselines


def get_baseline_indices(
    unique_baselines: np.ndarray, baseline_set: np.ndarray
) -> np.ndarray:
    """
    Finds the baseline indices of a set of baselines using the unique baselines.

    Uses a pairing function to reduce the number of values so it's more
    efficient to find the indices.

    Parameters
    ----------
    unique_baselines : np.ndarray
        a 2D array of unique antenna pairs (baselines).
    baseline_set : np.ndarray
        a 2D array of antenna pairs (baselines). This array may contain duplicates.

    Returns
    -------
    baseline_indices : np.ndarray
        the indices of the baseline set that
        correspond to the unique baselines.
    """
    unique_baselines_paired = pairing_function(unique_baselines)
    baseline_set_paired = pairing_function(baseline_set)

    # Pairing function doesn't preserve order so they need to be sorted.
    unique_baselines_sorted = np.argsort(unique_baselines_paired)
    sorted_indices = np.searchsorted(
        unique_baselines_paired[unique_baselines_sorted],
        baseline_set_paired,
    )
    baseline_indices = unique_baselines_sorted[sorted_indices]

    return baseline_indices



def add_encoding(xds, compressor, chunks=None):
    if chunks is None:
        chunks = xds.sizes

    chunks = {**dict(xds.sizes), **chunks}  # Add missing sizes if presents.

    for da_name in list(xds.data_vars):
        da_chunks = [chunks[dim_name] for dim_name in xds[da_name].sizes]
        xds[da_name].encoding = {"compressor": compressor, "chunks": da_chunks}



def make_quantity(value, units: str, dims: list = []) -> dict:
    """
    create a quantity dictionary given value and units

    Parameters
    ----------
    value : numeric or array of numerics
        Quantity value
    units: str
        Quantity units

    Returns
    -------
    dict
    """
    return {"data": value, "dims": dims, "attrs": make_quantity_attrs(units)}


def ensure_units_are_consistent(units):
    if isinstance(units, str):
        return units
    else:
        u0 = units[0]
        for u in units:
            assert u0 == u, f"Units are not consistent: {u0} != {u}. "
        return u0


def make_quantity_attrs(units: str) -> dict:
    """
    Creates the dict of attributes of a quantity

    Parameters
    ----------
    units: str
        Quantity units

    Returns
    -------
    dict
    """
    return {"units": ensure_units_are_consistent(units), "type": "quantity"}


def make_spectral_coord_reference_dict(
    value: float, units: str, observer: str = "lsrk"
) -> dict:
    """
    creates a spectral_coord measure dict given the value, units, and observer

    Parameters
    ----------
    value : numeric or array of numerics
        measure value
    units : str
        measure units
    observer :
        observer reference frame

    Returns
    -------
    dict
    """
    u = ensure_units_are_consistent(units)
    return {
        "attrs": make_spectral_coord_measure_attrs(
            u,
            observer.lower() if observer not in ["TOPO", "BARY", "REST"] else observer,
        ),
        "data": value,
        "dims": [],
    }


def make_spectral_coord_measure_attrs(units: str, observer: str = "lsrk") -> dict:
    """
    Creates a spectral_coord measure attrs dict given units and observer

    Parameters
    ----------
    units: str or list of str
        Spectral coordinate units
    observer: str
        Spectral reference frame
    Returns
    -------
    dict
        Attrs dict for a spectral_coord measure
    """
    u = ensure_units_are_consistent(units)
    return {"units": u, "observer": observer, "type": "spectral_coord"}


def make_skycoord_dict(data: list[float], units: str, frame: str) -> dict:
    return {
        "attrs": {
            "frame": frame.lower(),
            "type": "sky_coord",
            "units": ensure_units_are_consistent(units),
        },
        "data": data,
        "dims": ["l", "m"],
    }


def make_time_measure_attrs(units="s", scale="utc", time_format="mjd") -> dict:
    u = ensure_units_are_consistent(units)
    return {"units": u, "scale": scale, "format": time_format, "type": "time"}


def make_time_measure_dict(data, units="s", scale="utc", time_format="mjd") -> dict:
    """
    create a time measure dictionary given value and units
    Parameters
    ----------
    value : numeric or array of numerics
        Time value
    units: str
        Time units
    scale: str
        Time scale
    time_format: str
        Time format
    Returns
    -------
    dict
    """
    x = {}
    x["attrs"] = make_time_measure_attrs(units, scale, time_format)
    x["data"] = data
    x["dims"] = []
    return x


def make_time_coord_attrs(units="s", scale="utc", time_format="mjd") -> dict:
    """
    create a time measure dictionary given value and units
    Parameters
    ----------
    value : numeric or array of numerics
        Time value
    units: str
        Time units
    scale: str
        Time scale
    time_format: str
        Time format
    Returns
    -------
    dict
    """
    x = make_time_measure_attrs(units, scale, time_format)
    del x["type"]
    return x


def _casacore_q_to_xradio_q(q: dict) -> dict:
    """
    Convert a casacore quantity to an xradio quantity
    """
    if isinstance(q, dict):
        if "value" in q and "unit" in q:
            return make_quantity(q["value"], q["unit"])
        else:
            p = {}
            for k in q:
                p[k] = _casacore_q_to_xradio_q(q[k])
            return p
    else:
        raise ValueError(f"Cannot convert {q} to xradio quantity")



import pandas as pd
from xradio._utils.list_and_array import to_list
import numpy as np
import xarray as xr

PS_DATASET_TYPES = {"processing_set"}


class InvalidAccessorLocation(ValueError):
    """
    Raised by Processing Set accessor functions called on a wrong DataTree node (not processing set).
    """

    pass


@xr.register_datatree_accessor("xr_ps")
class ProcessingSetXdt:
    """
    Accessor to Processing Set DataTree nodes. Provides Processing Set specific functionality such
    as producing a summary of the processing set (with information from all its MSv4s), or retrieving
    combined antenna or field_and_source datasets.
    """

    _xdt: xr.DataTree

    def __init__(self, datatree: xr.DataTree):
        """
        Initialize the ProcessingSetXdt instance.

        Parameters
        ----------
        datatree: xarray.DataTree
            The Processing Set DataTree node to construct a ProcessingSetXdt accessor.
        """

        self._xdt = datatree
        self.meta = {"summary": {}}

    def summary(self, data_group: str = None) -> pd.DataFrame:
        """
        Generate and retrieve a summary of the Processing Set.

        The summary includes information such as the names of the Measurement Sets,
        their intents, polarizations, spectral window names, field names, source names,
        field coordinates, start frequencies, and end frequencies.

        Parameters
        ----------
        data_group : str, optional
            The data group to summarize. By default the "base" group
            is used (if found), or otherwise the first group found.

        Returns
        -------
        pandas.DataFrame
            A DataFrame containing the summary information of the specified data group.
        """

        def find_data_group_base_or_first(data_group: str, xdt: xr.DataTree) -> str:
            first_msv4 = next(iter(xdt.values()))
            first_data_groups = first_msv4.attrs["data_groups"]
            if data_group is None:
                data_group = (
                    "base"
                    if "base" in first_data_groups
                    else next(iter(first_data_groups))
                )
            return data_group

        if self._xdt.attrs.get("type") not in PS_DATASET_TYPES:
            raise InvalidAccessorLocation(
                f"{self._xdt.path} is not a processing set node."
            )

        data_group = find_data_group_base_or_first(data_group, self._xdt)

        if data_group in self.meta["summary"]:
            return self.meta["summary"][data_group]
        else:
            self.meta["summary"][data_group] = self._summary(data_group).sort_values(
                by=["name"], ascending=True
            )
            return self.meta["summary"][data_group]

    def get_max_dims(self) -> dict[str, int]:
        """
        Determine the maximum dimensions across all Measurement Sets in the Processing Set.

        This method examines each Measurement Set's dimensions and computes the maximum
        size for each dimension across the entire Processing Set.

        For example, if the Processing Set contains two MSs with dimensions (50, 20, 30) and (10, 30, 40),
        the maximum dimensions will be (50, 30, 40).

        Returns
        -------
        dict
            A dictionary containing the maximum dimensions of the Processing Set, with dimension names as keys
            and their maximum sizes as values.
        """

        if self._xdt.attrs.get("type") not in PS_DATASET_TYPES:
            raise InvalidAccessorLocation(
                f"{self._xdt.path} is not a processing set node."
            )

        if "max_dims" in self.meta:
            return self.meta["max_dims"]
        else:
            max_dims = None
            for ms_xdt in self._xdt.values():
                if max_dims is None:
                    max_dims = dict(ms_xdt.sizes)
                else:
                    for dim_name, size in ms_xdt.sizes.items():
                        if dim_name in max_dims:
                            if max_dims[dim_name] < size:
                                max_dims[dim_name] = size
                        else:
                            max_dims[dim_name] = size
            self.meta["max_dims"] = max_dims
            return self.meta["max_dims"]

    def get_freq_axis(self) -> xr.DataArray:
        """
        Combine the frequency axes of all Measurement Sets in the Processing Set.

        This method aggregates the frequency information from each Measurement Set to create
        a unified frequency axis for the entire Processing Set.

        Returns
        -------
        xarray.DataArray
            The combined frequency axis of the Processing Set.
        """
        if self._xdt.attrs.get("type") not in PS_DATASET_TYPES:
            raise InvalidAccessorLocation(
                f"{self._xdt.path} is not a processing set node."
            )

        if "freq_axis" in self.meta:
            return self.meta["freq_axis"]
        else:
            spw_names = []
            freq_axis_list = []
            frame = self._xdt[next(iter(self._xdt.children))].frequency.attrs[
                "observer"
            ]
            for ms_xdt in self._xdt.values():
                assert (
                    frame == ms_xdt.frequency.attrs["observer"]
                ), "Frequency reference frame not consistent in Processing Set."
                if ms_xdt.frequency.attrs["spectral_window_name"] not in spw_names:
                    spw_names.append(ms_xdt.frequency.attrs["spectral_window_name"])
                    freq_axis_list.append(ms_xdt.frequency)

            freq_axis = xr.concat(freq_axis_list, dim="frequency", join="outer").sortby(
                "frequency"
            )
            self.meta["freq_axis"] = freq_axis
            return self.meta["freq_axis"]

    def _summary(self, data_group: str = None):
        summary_data = {
            "name": [],
            "intents": [],
            "shape": [],
            "polarization": [],
            "scan_name": [],
            "spw_name": [],
            "spw_intent": [],
            "field_name": [],
            "source_name": [],
            "line_name": [],
            "field_coords": [],
            "start_frequency": [],
            "end_frequency": [],
        }
        from astropy.coordinates import SkyCoord
        import astropy.units as u

        for key, value in self._xdt.items():
            partition_info = value.xr_ms.get_partition_info()

            summary_data["name"].append(key)
            summary_data["intents"].append(partition_info["intents"])
            summary_data["spw_name"].append(partition_info["spectral_window_name"])
            summary_data["spw_intent"].append(partition_info["spectral_window_intent"])
            summary_data["polarization"].append(value.polarization.values)
            summary_data["scan_name"].append(partition_info["scan_name"])
            data_name = value.attrs["data_groups"][data_group]["correlated_data"]

            if "VISIBILITY" in data_name:
                center_name = "FIELD_PHASE_CENTER_DIRECTION"

            if "SPECTRUM" in data_name:
                center_name = "FIELD_REFERENCE_CENTER_DIRECTION"

            summary_data["shape"].append(value[data_name].shape)

            summary_data["field_name"].append(partition_info["field_name"])
            summary_data["source_name"].append(partition_info["source_name"])

            summary_data["line_name"].append(partition_info["line_name"])

            summary_data["start_frequency"].append(
                to_list(value["frequency"].values)[0]
            )
            summary_data["end_frequency"].append(to_list(value["frequency"].values)[-1])

            field_and_source_xds = value.xr_ms.get_field_and_source_xds(data_group)

            if field_and_source_xds.attrs["type"] == "field_and_source_ephemeris":
                summary_data["field_coords"].append("Ephemeris")
            elif field_and_source_xds[center_name]["field_name"].size > 1:
                summary_data["field_coords"].append("Multi-Phase-Center")
            else:
                ra_dec_rad = field_and_source_xds[center_name].values[0, :]
                frame = field_and_source_xds[center_name].attrs["frame"].lower()

                coord = SkyCoord(
                    ra=ra_dec_rad[0] * u.rad, dec=ra_dec_rad[1] * u.rad, frame=frame
                )

                summary_data["field_coords"].append(
                    [
                        frame,
                        coord.ra.to_string(unit=u.hour, precision=2),
                        coord.dec.to_string(unit=u.deg, precision=2),
                    ]
                )

        summary_df = pd.DataFrame(summary_data)
        return summary_df

    def query(
        self, string_exact_match: bool = True, query: str = None, **kwargs
    ) -> xr.DataTree:
        """
        Select a subset of the Processing Set based on specified criteria.

        This method allows filtering the Processing Set by matching column names and values
        or by applying a Pandas query string. The selection criteria can target various
        attributes of the Measurement Sets such as intents, polarization, spectral window names, etc.

        A data group can be selected by name by using the `data_group_name` parameter. This is applied to each Measurement Set in the Processing Set.

        Note
        ----
        This selection does not modify the actual data within the Measurement Sets. For example, if
        a Measurement Set has `field_name=['field_0','field_10','field_08']` and `ps.query(field_name='field_0')`
        is invoked, the resulting subset will still contain the original list `['field_0','field_10','field_08']`.
        The exception is data group selection, using `data_group_name`, that will select data variables only associated with the specified data group in the Measurement Set.

        Parameters
        ----------
        string_exact_match : bool, optional
            If `True`, string matching will require exact matches for string and string list columns.
            If `False`, partial matches are allowed. Default is `True`.
        query : str, optional
            A Pandas query string to apply additional filtering. Default is `None`.
        **kwargs : dict
            Keyword arguments representing column names and their corresponding values to filter the Processing Set.

        Returns
        -------
        xr.DataTree
            A new Processing Set DataTree instance containing only the Measurement Sets that match the selection criteria.

        Examples
        --------
        >>> # Select all MSs with intents 'OBSERVE_TARGET#ON_SOURCE' and polarization 'RR' or 'LL'
        >>> selected_ps_xdt = ps_xdt.xr_ps.query(intents='OBSERVE_TARGET#ON_SOURCE', polarization=['RR', 'LL'])

        >>> # Select all MSs with start_frequency greater than 100 GHz and less than 200 GHz
        >>> selected_ps_xdt = ps_xdt.xr_ps.query(query='start_frequency > 100e9 AND end_frequency < 200e9')
        """

        if self._xdt.attrs.get("type") not in PS_DATASET_TYPES:
            raise InvalidAccessorLocation(
                f"{self._xdt.path} is not a processing set node."
            )

        def select_rows(df, col, sel_vals, string_exact_match):
            def check_selection(row_val):
                row_val = to_list(
                    row_val
                )  # make sure that it is a list so that we can iterate over it.

                for rw in row_val:
                    for s in sel_vals:
                        if string_exact_match:
                            if rw == s:
                                return True
                        else:
                            if s in rw:
                                return True
                return False

            return df[df[col].apply(check_selection)]

        summary_table = self.summary()
        data_group_name = None
        for key, value in kwargs.items():

            if "data_group_name" == key:
                data_group_name = value
            else:
                value = to_list(value)  # make sure value is a list.

                if len(value) == 1 and isinstance(value[0], slice):
                    summary_table = summary_table[
                        summary_table[key].between(value[0].start, value[0].stop)
                    ]
                else:
                    summary_table = select_rows(
                        summary_table, key, value, string_exact_match
                    )

        if query is not None:
            summary_table = summary_table.query(query)

        sub_ps_xdt = xr.DataTree()
        for key, val in self._xdt.items():
            if key in summary_table["name"].values:
                if data_group_name is not None:
                    sub_ps_xdt[key] = val.xr_ms.sel(data_group_name=data_group_name)
                else:
                    sub_ps_xdt[key] = val

        sub_ps_xdt.attrs = self._xdt.attrs

        return sub_ps_xdt

    def get_combined_field_and_source_xds(self, data_group: str = "base") -> xr.Dataset:
        """
        Combine all non-ephemeris `field_and_source_xds` datasets from a Processing Set for a data group into a
        single dataset.

        Parameters
        ----------
        data_group : str, optional
            The data group to process. Default is "base".

        Returns
        -------
        xarray.Dataset
            combined_field_and_source_xds: Combined dataset for standard (non-ephemeris) fields.

        Raises
        ------
        ValueError
            If the `field_and_source_xds` attribute is missing or improperly formatted in any Measurement Set.
        """

        if self._xdt.attrs.get("type") not in PS_DATASET_TYPES:
            raise InvalidAccessorLocation(
                f"{self._xdt.path} is not a processing set node."
            )

        combined_field_and_source_xds = xr.Dataset()
        for ms_name, ms_xdt in self._xdt.items():
            field_and_source_xds = ms_xdt.xr_ms.get_field_and_source_xds(data_group)

            if not field_and_source_xds.attrs["type"] == "field_and_source_ephemeris":

                if (
                    "line_name" in field_and_source_xds.coords
                ):  # Not including line info since it is a function of spw.
                    field_and_source_xds = field_and_source_xds.drop_vars(
                        ["LINE_REST_FREQUENCY", "LINE_SYSTEMIC_VELOCITY"],
                        errors="ignore",
                    )
                    del field_and_source_xds["line_name"]
                    del field_and_source_xds["line_label"]

                if len(combined_field_and_source_xds.data_vars) == 0:
                    combined_field_and_source_xds = field_and_source_xds
                else:
                    combined_field_and_source_xds = xr.concat(
                        [combined_field_and_source_xds, field_and_source_xds],
                        dim="field_name",
                        join="outer",
                    )

        if (len(combined_field_and_source_xds.data_vars) > 0) and (
            "FIELD_PHASE_CENTER_DIRECTION" in combined_field_and_source_xds
        ):
            combined_field_and_source_xds = (
                combined_field_and_source_xds.drop_duplicates("field_name")
            )

            combined_field_and_source_xds["MEAN_PHASE_CENTER_DIRECTION"] = (
                combined_field_and_source_xds["FIELD_PHASE_CENTER_DIRECTION"].mean(
                    dim=["field_name"]
                )
            )

            ra1 = (
                combined_field_and_source_xds["FIELD_PHASE_CENTER_DIRECTION"]
                .sel(sky_dir_label="ra")
                .values
            )
            dec1 = (
                combined_field_and_source_xds["FIELD_PHASE_CENTER_DIRECTION"]
                .sel(sky_dir_label="dec")
                .values
            )
            ra2 = (
                combined_field_and_source_xds["MEAN_PHASE_CENTER_DIRECTION"]
                .sel(sky_dir_label="ra")
                .values
            )
            dec2 = (
                combined_field_and_source_xds["MEAN_PHASE_CENTER_DIRECTION"]
                .sel(sky_dir_label="dec")
                .values
            )

            from xradio._utils.coord_math import haversine

            distance = haversine(ra1, dec1, ra2, dec2)
            min_index = distance.argmin()

            combined_field_and_source_xds.attrs["center_field_name"] = (
                combined_field_and_source_xds.field_name[min_index].values
            )

        return combined_field_and_source_xds

    def get_combined_field_and_source_xds_ephemeris(
        self, data_group: str = "base"
    ) -> xr.Dataset:
        """
        Combine all ephemeris `field_and_source_xds` datasets from a Processing Set for a datagroup into a single dataset.

        Parameters
        ----------
        data_group : str, optional
            The data group to process. Default is "base".

        Returns
        -------
        xarray.Dataset
            combined_ephemeris_field_and_source_xds: Combined dataset for ephemeris fields.

        Raises
        ------
        ValueError
            If the `field_and_source_xds` attribute is missing or improperly formatted in any Measurement Set.
        """

        if self._xdt.attrs.get("type") not in PS_DATASET_TYPES:
            raise InvalidAccessorLocation(
                f"{self._xdt.path} is not a processing set node."
            )

        combined_ephemeris_field_and_source_xds = xr.Dataset()
        for ms_name, ms_xdt in self._xdt.items():
            field_and_source_xds = field_and_source_xds = (
                ms_xdt.xr_ms.get_field_and_source_xds(data_group)
            )

            if field_and_source_xds.attrs["type"] == "field_and_source_ephemeris":

                if (
                    "line_name" in field_and_source_xds.coords
                ):  # Not including line info since it is a function of spw.
                    field_and_source_xds = field_and_source_xds.drop_vars(
                        ["LINE_REST_FREQUENCY", "LINE_SYSTEMIC_VELOCITY"],
                        errors="ignore",
                    )
                    del field_and_source_xds["line_name"]
                    del field_and_source_xds["line_label"]

                from xradio.measurement_set._utils._utils.interpolate import (
                    interpolate_to_time,
                )

                if "time_ephemeris" in field_and_source_xds:
                    field_and_source_xds = interpolate_to_time(
                        field_and_source_xds,
                        field_and_source_xds.time,
                        "field_and_source_xds",
                        "time_ephemeris",
                    )
                    del field_and_source_xds["time_ephemeris"]
                    field_and_source_xds = field_and_source_xds.rename(
                        {"time_ephemeris": "time"}
                    )

                if "OBSERVER_POSITION" in field_and_source_xds:
                    field_and_source_xds = field_and_source_xds.drop_vars(
                        ["OBSERVER_POSITION"], errors="ignore"
                    )

                if len(combined_ephemeris_field_and_source_xds.data_vars) == 0:
                    combined_ephemeris_field_and_source_xds = field_and_source_xds
                else:

                    combined_ephemeris_field_and_source_xds = xr.concat(
                        [combined_ephemeris_field_and_source_xds, field_and_source_xds],
                        dim="time",
                        join="outer",
                    )

        if (len(combined_ephemeris_field_and_source_xds.data_vars) > 0) and (
            "FIELD_PHASE_CENTER_DIRECTION" in combined_ephemeris_field_and_source_xds
        ):

            from xradio._utils.coord_math import wrap_to_pi

            offset = (
                combined_ephemeris_field_and_source_xds["FIELD_PHASE_CENTER_DIRECTION"]
                - combined_ephemeris_field_and_source_xds["SOURCE_DIRECTION"]
            )
            combined_ephemeris_field_and_source_xds["FIELD_OFFSET"] = xr.DataArray(
                wrap_to_pi(offset.sel(sky_dir_label=["ra", "dec"])).values,
                dims=["time", "sky_dir_label"],
            )
            combined_ephemeris_field_and_source_xds["FIELD_OFFSET"].attrs = (
                combined_ephemeris_field_and_source_xds[
                    "FIELD_PHASE_CENTER_DIRECTION"
                ].attrs
            )
            combined_ephemeris_field_and_source_xds["FIELD_OFFSET"].attrs["units"] = (
                combined_ephemeris_field_and_source_xds["FIELD_OFFSET"].attrs["units"][
                    :2
                ]
            )

            ra1 = (
                combined_ephemeris_field_and_source_xds["FIELD_OFFSET"]
                .sel(sky_dir_label="ra")
                .values
            )
            dec1 = (
                combined_ephemeris_field_and_source_xds["FIELD_OFFSET"]
                .sel(sky_dir_label="dec")
                .values
            )
            ra2 = 0.0
            dec2 = 0.0

            from xradio._utils.coord_math import haversine

            distance = haversine(ra1, dec1, ra2, dec2)
            min_index = distance.argmin()

            combined_ephemeris_field_and_source_xds.attrs["center_field_name"] = (
                combined_ephemeris_field_and_source_xds.field_name[min_index].values
            )

        return combined_ephemeris_field_and_source_xds

    def plot_phase_centers(
        self, label_all_fields: bool = False, data_group: str = "base"
    ):
        """
        Plot the phase center locations of all fields in the Processing Set.

        This method is primarily used for visualizing mosaics. It generates scatter plots of
        the phase center coordinates for both standard and ephemeris fields. The central field
        is highlighted in red based on the closest phase center calculation.

        Parameters
        ----------
        label_all_fields : bool, optional
            If `True`, all fields will be labeled on the plot. Default is `False`.
        data_group : str, optional
            The data group to use for processing. Default is "base".

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the combined datasets are empty or improperly formatted.
        """

        def setup_annotations_all(axis, scatter, field_names):
            """
            Creates annotations for when label_all_fields=True
            """
            coord_x, coord_y = np.array(scatter.get_offsets()).transpose()
            offset_x = np.abs(np.max(coord_x) - np.min(coord_x)) * 0.01
            offset_y = np.abs(np.max(coord_y) - np.min(coord_y)) * 0.01
            for idx, (x, y) in enumerate(zip(coord_x + offset_x, coord_y + offset_y)):
                axis.annotate(field_names[idx], (x, y), alpha=1)

        if self._xdt.attrs.get("type") not in PS_DATASET_TYPES:
            raise InvalidAccessorLocation(
                f"{self._xdt.path} is not a processing set node."
            )

        combined_field_and_source_xds = self.get_combined_field_and_source_xds(
            data_group
        )
        combined_ephemeris_field_and_source_xds = (
            self.get_combined_field_and_source_xds_ephemeris(data_group)
        )
        from matplotlib import pyplot as plt

        if (len(combined_field_and_source_xds.data_vars) > 0) and (
            "FIELD_PHASE_CENTER_DIRECTION" in combined_field_and_source_xds
        ):
            fig = plt.figure()
            plt.title("Field Phase Center Locations")
            scatter = plt.scatter(
                combined_field_and_source_xds["FIELD_PHASE_CENTER_DIRECTION"].sel(
                    sky_dir_label="ra"
                ),
                combined_field_and_source_xds["FIELD_PHASE_CENTER_DIRECTION"].sel(
                    sky_dir_label="dec"
                ),
            )
            center_field_name = combined_field_and_source_xds.attrs["center_field_name"]
            center_field = combined_field_and_source_xds.sel(
                field_name=center_field_name
            )

            if label_all_fields:
                field_name = combined_field_and_source_xds.field_name.values
                setup_annotations_all(fig.axes[0], scatter, field_name)
                fig.axes[0].margins(0.2, 0.2)
                center_label = None
            else:
                center_label = center_field_name

            plt.scatter(
                center_field["FIELD_PHASE_CENTER_DIRECTION"].sel(sky_dir_label="ra"),
                center_field["FIELD_PHASE_CENTER_DIRECTION"].sel(sky_dir_label="dec"),
                color="red",
                label=center_label,
            )
            plt.xlabel("RA (rad)")
            plt.ylabel("DEC (rad)")
            if not label_all_fields:
                plt.legend()
            plt.show()

        if (len(combined_ephemeris_field_and_source_xds.data_vars) > 0) and (
            "FIELD_PHASE_CENTER_DIRECTION" in combined_ephemeris_field_and_source_xds
        ):

            fig = plt.figure()
            plt.title(
                "Offset of Field Phase Center from Source Location (Ephemeris Data)"
            )
            scatter = plt.scatter(
                combined_ephemeris_field_and_source_xds["FIELD_OFFSET"].sel(
                    sky_dir_label="ra"
                ),
                combined_ephemeris_field_and_source_xds["FIELD_OFFSET"].sel(
                    sky_dir_label="dec"
                ),
            )

            center_field_name = combined_ephemeris_field_and_source_xds.attrs[
                "center_field_name"
            ]

            combined_ephemeris_field_and_source_xds = (
                combined_ephemeris_field_and_source_xds.set_xindex("field_name")
            )

            center_field = combined_ephemeris_field_and_source_xds.sel(
                field_name=center_field_name
            )

            if label_all_fields:
                field_name = combined_ephemeris_field_and_source_xds.field_name.values
                setup_annotations_all(fig.axes[0], scatter, field_name)
                fig.axes[0].margins(0.2, 0.2)
                center_label = None
            else:
                center_label = center_field_name

            plt.scatter(
                center_field["FIELD_OFFSET"].sel(sky_dir_label="ra"),
                center_field["FIELD_OFFSET"].sel(sky_dir_label="dec"),
                color="red",
                label=center_label,
            )

            plt.xlabel("RA Offset (rad)")
            plt.ylabel("DEC Offset (rad)")
            if not label_all_fields:
                plt.legend()
            plt.show()

    def get_combined_antenna_xds(self) -> xr.Dataset:
        """
        Combine the `antenna_xds` datasets from all Measurement Sets into a single dataset.

        This method concatenates the antenna datasets from each Measurement Set along the 'antenna_name' dimension.

        Returns
        -------
        xarray.Dataset
            A combined `xarray.Dataset` containing antenna information from all Measurement Sets.

        Raises
        ------
        ValueError
            If antenna datasets are missing required variables or improperly formatted.
        """
        if self._xdt.attrs.get("type") not in PS_DATASET_TYPES:
            raise InvalidAccessorLocation(
                f"{self._xdt.path} is not a processing set node."
            )

        combined_antenna_xds = xr.Dataset()
        for cor_name, ms_xdt in self._xdt.items():
            antenna_xds = ms_xdt.antenna_xds.ds

            if len(combined_antenna_xds.data_vars) == 0:
                combined_antenna_xds = antenna_xds
            else:
                combined_antenna_xds = xr.concat(
                    [combined_antenna_xds, antenna_xds],
                    dim="antenna_name",
                    data_vars="minimal",
                    coords="minimal",
                    join="outer",
                )

        # ALMA WVR antenna_xds data has a NaN value for the antenna receptor angle.
        if "ANTENNA_RECEPTOR_ANGLE" in combined_antenna_xds.data_vars:
            combined_antenna_xds = combined_antenna_xds.dropna("antenna_name")

        combined_antenna_xds = combined_antenna_xds.drop_duplicates("antenna_name")

        return combined_antenna_xds

    def plot_antenna_positions(self, label_all_antennas: bool = False):
        """
        Plot the antenna positions of all antennas in the Processing Set.

        This method generates and displays a figure with three scatter plots, displaying the antenna
        positions in different planes:

        - X vs Y
        - X vs Z
        - Y vs Z

        The antenna names are shown on hovering their positions, unless label_all_antennas is enabled.

        Parameters
        ----------
        label_all_antennas : bool, optional
            If 'True', annotations are shown with the names of every antenna next to their positions.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the combined antenna dataset is empty or missing required coordinates.
        """

        def antenna_hover(event):
            if event.inaxes in antenna_axes:
                for axis in antenna_axes:
                    contained, indices = scatter_map[axis].contains(event)
                    annotation = annotations_map[axis]
                    if contained:
                        scatter = scatter_map[axis]
                        update_antenna_annotation(indices, scatter, annotation)
                        annotation.set_visible(True)
                        fig.canvas.draw_idle()
                    else:
                        visible = annotation.get_visible()
                        if visible:
                            annotation.set_visible(False)
                            fig.canvas.draw_idle()

        def update_antenna_annotation(indices, scatter, annotation):
            position = scatter.get_offsets()[indices["ind"][0]]
            annotation.xy = position
            text = "{}".format(" ".join([antenna_names[num] for num in indices["ind"]]))
            annotation.set_text(text)
            annotation.get_bbox_patch().set_facecolor("#e8d192")
            annotation.get_bbox_patch().set_alpha(1)

        def setup_annotations_for_hover(antenna_axes, scatter_plots):
            """
            Creates annotations on all the axes requested.

            Returns
            -------
            dict
                dict from antenna axes -> annotation objects
            """
            antenna_annotations = []
            for axis in antenna_axes:
                annotation = axis.annotate(
                    "",
                    xy=(0, 0),
                    xytext=(10, 15),
                    textcoords="offset points",
                    arrowprops=dict(arrowstyle="-|>"),
                    bbox=dict(boxstyle="round", fc="w"),
                )
                antenna_annotations.append(annotation)
                annotation.set_visible(False)
            annotations_map = dict(zip(antenna_axes, antenna_annotations))

            return annotations_map

        def setup_annotations_for_all(antenna_axes, scatter_map):
            """
            Creates annotations for when label_all_antennas=True
            """
            for axis in antenna_axes:
                scatter = scatter_map[axis]
                coord_x, coord_y = np.array(scatter.get_offsets()).transpose()
                offset_x = np.abs(np.max(coord_x) - np.min(coord_x)) * 0.01
                offset_y = np.abs(np.max(coord_y) - np.min(coord_y)) * 0.01
                for idx, (x, y) in enumerate(
                    zip(coord_x + offset_x, coord_y + offset_y)
                ):
                    axis.annotate(
                        antenna_names[idx],
                        (x, y),
                        alpha=1,
                    )

        if self._xdt.attrs.get("type") not in PS_DATASET_TYPES:
            raise InvalidAccessorLocation(
                f"{self._xdt.path} is not a processing set node."
            )

        combined_antenna_xds = self.get_combined_antenna_xds()
        from matplotlib import pyplot as plt

        fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(12, 8))
        fig.suptitle("Antenna Positions")
        fig.subplots_adjust(
            wspace=0.25, hspace=0.25, left=0.1, right=0.95, top=0.9, bottom=0.1
        )

        scatter1 = ax1.scatter(
            combined_antenna_xds["ANTENNA_POSITION"].sel(cartesian_pos_label="x"),
            combined_antenna_xds["ANTENNA_POSITION"].sel(cartesian_pos_label="y"),
        )
        ax1.set_xlabel("x (m)")
        ax1.set_ylabel("y (m)")
        antenna_names = combined_antenna_xds.antenna_name.values

        scatter2 = ax2.scatter(
            combined_antenna_xds["ANTENNA_POSITION"].sel(cartesian_pos_label="y"),
            combined_antenna_xds["ANTENNA_POSITION"].sel(cartesian_pos_label="z"),
        )
        ax2.set_xlabel("y (m)")
        ax2.set_ylabel("z (m)")

        scatter3 = ax3.scatter(
            combined_antenna_xds["ANTENNA_POSITION"].sel(cartesian_pos_label="x"),
            combined_antenna_xds["ANTENNA_POSITION"].sel(cartesian_pos_label="z"),
        )
        ax3.set_xlabel("x (m)")
        ax3.set_ylabel("z (m)")

        ax4.axis("off")

        antenna_axes = [ax1, ax2, ax3]
        scatter_map = dict(zip(antenna_axes, [scatter1, scatter2, scatter3]))
        if label_all_antennas:
            annotations_map = setup_annotations_for_all(antenna_axes, scatter_map)
        else:
            annotations_map = setup_annotations_for_hover(antenna_axes, scatter_map)
            fig.canvas.mpl_connect("motion_notify_event", antenna_hover)

        plt.show()



import copy
import re
from typing import Union

import dask
import dask.array as da
import numpy as np
import psutil
import xarray as xr
from astropy import units as u
from astropy.io import fits
from astropy.time import Time

from xradio._utils.coord_math import _deg_to_rad
from xradio._utils.dict_helpers import (
    make_quantity,
    make_spectral_coord_reference_dict,
    make_skycoord_dict,
    make_time_measure_dict,
)

from xradio.measurement_set._utils._utils.stokes_types import stokes_types
from xradio.image._util.common import (
    _compute_linear_world_values,
    _compute_velocity_values,
    _compute_world_sph_dims,
    _convert_beam_to_rad,
    _default_freq_info,
    _doppler_types,
    _freq_from_vel,
    _get_unit,
    _get_xds_dim_order,
    _image_type,
    _l_m_attr_notes,
)


def _fits_image_to_xds(
    img_full_path: str,
    chunks: dict,
    verbose: bool,
    do_sky_coords: bool,
    compute_mask: bool,
) -> dict:
    """
    compute_mask : bool, optional
        If True (default), compute and attach valid data masks to the xds.
        If False, skip mask generation for performance. It is solely the responsibility
        of the user to ensure downstream apps can handle NaN values; do not
        ask package developers to add this non-standard behavior.

    TODO: complete documentation
    Create an xds without any pixel data from metadata from the specified FITS image
    """
    # may also need to pass mode='denywrite'
    # https://stackoverflow.com/questions/35759713/astropy-io-fits-read-row-from-large-fits-file-with-mutliple-hdus
    try:
        hdulist = fits.open(img_full_path, memmap=True)
        attrs, helpers, header = _fits_header_to_xds_attrs(hdulist, compute_mask)
    finally:
        hdulist.close()
        # avoid keeping reference to mem-mapped fits file
        del hdulist
    xds = _create_coords(helpers, header, do_sky_coords)
    sphr_dims = helpers["sphr_dims"]
    ary = _read_image_array(img_full_path, chunks, helpers, verbose)
    dim_order = _get_xds_dim_order(sphr_dims)
    xds = _add_sky_or_aperture(xds, ary, dim_order, header, helpers, sphr_dims)
    xds.attrs = attrs
    xds = _add_coord_attrs(xds, helpers)
    if helpers["has_multibeam"]:
        xds = _do_multibeam(xds, img_full_path)
    elif "beam" in helpers and helpers["beam"] is not None:
        xds = _add_beam(xds, helpers)
    return xds


def _add_coord_attrs(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    xds = _add_time_attrs(xds, helpers)
    xds = _add_freq_attrs(xds, helpers)
    xds = _add_vel_attrs(xds, helpers)
    xds = _add_l_m_attrs(xds, helpers)
    xds = _add_lin_attrs(xds, helpers)
    return xds


def _add_time_attrs(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    time_coord = xds.coords["time"]
    time_coord.attrs = copy.deepcopy(helpers["obsdate"]["attrs"])
    xds.assign_coords(time=time_coord)
    return xds


def _add_freq_attrs(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    freq_coord = xds.coords["frequency"]
    meta = {}
    if helpers["has_freq"]:
        meta["rest_frequency"] = make_quantity(helpers["restfreq"], "Hz")
        meta["rest_frequencies"] = [meta["rest_frequency"]]
        meta["type"] = "frequency"
        meta["wave_units"] = "mm"
        freq_axis = helpers["freq_axis"]
        meta["reference_frequency"] = make_spectral_coord_reference_dict(
            helpers["crval"][freq_axis], "Hz", helpers["specsys"]
        )
        # meta["cdelt"] = helpers["cdelt"][freq_axis]
    if not meta:
        # this is the default frequency information CASA creates
        meta = _default_freq_info()
    freq_coord.attrs = copy.deepcopy(meta)
    xds["frequency"] = freq_coord
    return xds


def _add_vel_attrs(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    vel_coord = xds.coords["velocity"]
    meta = {"units": "m/s"}
    if helpers["has_freq"]:
        meta["doppler_type"] = helpers.get("doppler", "RADIO")
    else:
        meta["doppler_type"] = _doppler_types[0]
    meta["type"] = "doppler"
    vel_coord.attrs = copy.deepcopy(meta)
    xds.coords["velocity"] = vel_coord
    return xds


def _add_l_m_attrs(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    attr_note = _l_m_attr_notes()
    for c in ["l", "m"]:
        if c in xds.coords:
            xds[c].attrs = {
                "note": attr_note[c],
            }
    return xds


def _add_lin_attrs(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    if not helpers["sphr_dims"]:
        for i, j in zip(helpers["dir_axes"], ("u", "v")):
            meta = {
                "units": "wavelengths",
                "crval": helpers["crval"][i],
                "cdelt": helpers["cdelt"][i],
            }
            xds.coords[j].attrs = meta
    return xds


def _is_freq_like(v: str) -> bool:
    return v.startswith("FREQ") or v == "VOPT" or v == "VRAD"


def _xds_direction_attrs_from_header(helpers: dict, header) -> dict:
    # helpers is modified in place, headers is not modified
    t_axes = helpers["t_axes"]
    p0 = header[f"CTYPE{t_axes[0]}"][-3:]
    p1 = header[f"CTYPE{t_axes[1]}"][-3:]
    if p0 != p1:
        raise RuntimeError(
            f"Projections for direction axes ({p0}, {p1}) differ, but they "
            "must be the same"
        )
    direction = {}
    direction["projection"] = p0
    helpers["projection"] = p0
    ref_sys = header["RADESYS"]
    ref_eqx = None if ref_sys.upper() == "ICRS" else header["EQUINOX"]
    if ref_sys == "FK5" and ref_eqx == 2000:
        ref_eqx = "J2000.0"
    helpers["ref_sys"] = ref_sys
    helpers["ref_eqx"] = ref_eqx
    # fits does not support conversion frames
    direction["reference"] = make_skycoord_dict([0.0, 0.0], units="rad", frame=ref_sys)
    dir_axes = helpers["dir_axes"]
    ddata = []
    dunits = []
    for i in dir_axes:
        x = helpers["crval"][i] * u.Unit(_get_unit(helpers["cunit"][i]))
        x = x.to("rad")
        ddata.append(x.value)
        # direction["reference"]["value"][i] = x.value
        x = helpers["cdelt"][i] * u.Unit(_get_unit(helpers["cunit"][i]))
        dunits.append("rad")
    direction["reference"] = make_skycoord_dict(ddata, units=dunits, frame=ref_sys)
    if ref_eqx is not None:
        direction["reference"]["attrs"]["equinox"] = ref_eqx.lower()
    direction["latpole"] = make_quantity(
        header["LATPOLE"] * _deg_to_rad, "rad", dims=["l", "m"]
    )
    direction["lonpole"] = make_quantity(
        header["LONPOLE"] * _deg_to_rad, "rad", dims=["l", "m"]
    )
    pc = np.zeros([2, 2])
    for i in (0, 1):
        for j in (0, 1):
            # dir_axes are now 0-based, but fits needs 1-based
            try:
                pc[i][j] = header[f"PC{dir_axes[i]+1}_{dir_axes[j]+1}"]
            except KeyError:
                try:
                    pc[i][j] = header[f"PC0{dir_axes[i]+1}_0{dir_axes[j]+1}"]
                except KeyError:
                    raise RuntimeError(
                        f"Could not find PC{dir_axes[i]+1}_{dir_axes[j]+1} or "
                        f"PC0{dir_axes[i]+1}_0{dir_axes[j]+1} in FITS header"
                    )
    direction["pc"] = pc
    # Is there really no fits header parameter for projection_parameters?
    direction["projection_parameters"] = np.array([0.0, 0.0])
    return direction


def _fits_header_c_values_to_metadata(helpers: dict, header) -> None:
    # The helpers dict is modified in place. header is not modified
    ctypes = []
    shape = []
    crval = []
    cdelt = []
    crpix = []
    cunit = []
    for i in range(1, helpers["naxes"] + 1):
        ax_type = header[f"CTYPE{i}"]
        ctypes.append(ax_type)
        shape.append(header[f"NAXIS{i}"])
        crval.append(header[f"CRVAL{i}"])
        cdelt.append(header[f"CDELT{i}"])
        # FITS 1-based to python 0-based
        crpix.append(header[f"CRPIX{i}"] - 1)
        cunit.append(header[f"CUNIT{i}"])
    helpers["ax_type"] = ax_type
    helpers["shape"] = shape
    helpers["ctype"] = ctypes
    helpers["crval"] = crval
    helpers["cdelt"] = cdelt
    helpers["crpix"] = crpix
    helpers["cunit"] = cunit


def _get_telescope_metadata(helpers: dict, header) -> dict:
    # The helpers dict is modified in place. header is not modified
    tel = {}
    tel["name"] = header["TELESCOP"]
    if "OBSGEO-x" in header:
        x = header["OBSGEO-X"]
        y = header["OBSGEO-Y"]
        z = header["OBSGEO-Z"]
        xyz = np.array([x, y, z])
        r = np.sqrt(np.sum(xyz * xyz))
        lat = np.arcsin(z / r)
        long = np.arctan2(y, x)
        tel["direction"] = {
            "attrs": {
                "coordinate_system": "geocentric",
                # I haven't seen a FITS keyword for reference frame of telescope posiiton
                "frame": "ITRF",
                "origin_object_name": "earth",
                "type": "location",
                "units": "rad",
            },
            "data": np.array([long, lat]),
            "dims": ["ellipsoid_dir_label"],
            "coords": {
                "ellipsoid_dir_label": {
                    "dims": ["ellipsoid_dir_label"],
                    "data": ["lon", "lat"],
                }
            },
        }
        tel["distance"] = {
            "attrs": {
                "coordinate_system": "geocentric",
                # I haven't seen a FITS keyword for reference frame of telescope posiiton
                "frame": "ITRF",
                "origin_object_name": "earth",
                "type": "location",
                "units": "m",
            },
            "data": np.array([r]),
            "dims": ["ellipsoid_dis_label"],
            "coords": {
                "ellipsoid_dis_label": {
                    "dims": ["ellipsoid_dis_label"],
                    "data": [
                        "dist",
                    ],
                }
            },
        }
    return tel


def _compute_pointing_center(helpers: dict, header) -> dict:
    # Neither helpers or header is modified
    t_axes = helpers["t_axes"]
    long_unit = header[f"CUNIT{t_axes[0]}"]
    lat_unit = header[f"CUNIT{t_axes[1]}"]
    unit = []
    for uu in [long_unit, lat_unit]:
        new_u = u.Unit(_get_unit(uu))
        unit.append(new_u)
    pc_long = float(header[f"CRVAL{t_axes[0]}"]) * unit[0]
    pc_lat = float(header[f"CRVAL{t_axes[1]}"]) * unit[1]
    pc_long = pc_long.to(u.rad).value
    pc_lat = pc_lat.to(u.rad).value
    return make_skycoord_dict([pc_long, pc_lat], units="rad", frame=helpers["ref_sys"])


def _user_attrs_from_header(header) -> dict:
    # header is not modified
    exclude = [
        "ALTRPIX",
        "ALTRVAL",
        "BITPIX",
        "BMAJ",
        "BMIN",
        "BPA",
        "BSCALE",
        "BTYPE",
        "BUNIT",
        "BZERO",
        "CASAMBM",
        "DATE",
        "DATE-OBS",
        "EQUINOX",
        "EXTEND",
        "HISTORY",
        "LATPOLE",
        "LONPOLE",
        "OBSERVER",
        "ORIGIN",
        "TELESCOP",
        "OBJECT",
        "RADESYS",
        "RESTFRQ",
        "SIMPLE",
        "SPECSYS",
        "TIMESYS",
        "VELREF",
    ]
    regex = r"|".join(
        [
            r"^NAXIS\d?$",
            r"^CRVAL\d$",
            r"^CRPIX\d$",
            r"^CTYPE\d$",
            r"^CDELT\d$",
            r"^CUNIT\d$",
            r"^OBSGEO-(X|Y|Z)$",
            r"^P(C|V)0?\d_0?\d",
        ]
    )
    user = {}
    for k, v in header.items():
        if not (re.search(regex, k) or k in exclude):
            user[k.lower()] = v
    return user


def _beam_attr_from_header(helpers: dict, header) -> Union[dict, str, None]:
    # The helpers dict is modified in place. header is not modified
    helpers["has_multibeam"] = False
    if "BMAJ" in header:
        # single global beam
        beam = {
            "bmaj": make_quantity(header["BMAJ"], "deg"),
            "bmin": make_quantity(header["BMIN"], "deg"),
            "pa": make_quantity(header["BPA"], "deg"),
        }
        return _convert_beam_to_rad(beam)
    elif "CASAMBM" in header and header["CASAMBM"]:
        # multi-beam
        helpers["has_multibeam"] = True
        return "mb"
    else:
        # no beam
        return None


def _create_dim_map(helpers: dict, header) -> dict:
    # The helpers dict is modified in place. header is not modified
    t_axes = np.array([0, 0])
    dim_map = {}
    helpers["has_freq"] = False
    # fits indexing starts at 1, not 0
    for i in range(1, helpers["naxes"] + 1):
        ax_type = header[f"CTYPE{i}"]
        if ax_type.startswith("RA-"):
            t_axes[0] = i
        elif ax_type.startswith("DEC-"):
            t_axes[1] = i
        elif ax_type == "STOKES":
            dim_map["polarization"] = i - 1
        elif _is_freq_like(ax_type):
            dim_map["freq"] = i - 1
            helpers["has_freq"] = True
            # helpers["native_type"] = ax_type
        else:
            raise RuntimeError(f"{ax_type} is an unsupported axis")
    helpers["t_axes"] = t_axes
    helpers["dim_map"] = dim_map
    return dim_map


def _fits_header_to_xds_attrs(
    hdulist: fits.hdu.hdulist.HDUList, compute_mask: bool
) -> tuple:
    # First: Guard for unsupported compressed images
    for i, hdu in enumerate(hdulist):
        if isinstance(hdu, fits.CompImageHDU):
            raise RuntimeError(
                f"HDU {i}, name={hdu.name} is a CompImageHDU, which is not supported "
                "for memory-mapping. "
                "Cannot memory-map compressed FITS image (CompImageHDU). "
                "Workaround: decompress the FITS using tools like `funpack`, `cfitsio`, "
                "or Astropy's `.scale()`/`.copy()` workflows"
            )
    primary = None
    # FIXME beams is set but never actually used in this function. What's up with that?
    beams = None
    for hdu in hdulist:
        if hdu.name == "PRIMARY":
            primary = hdu
            # Memory map support check
            # avoid possibly non-existent hdu.scale_type attribute check and check header instead
            header = hdu.header
            scale = hdu.header.get("BSCALE", 1.0)
            zero = hdu.header.get("BZERO", 0.0)
            if not (scale == 1.0 and zero == 0.0):
                raise RuntimeError(
                    "Cannot memory-map scaled FITS data (BSCALE/BZERO set). "
                    f"BZERO={zero}, BSCALE={scale}. "
                    "Workaround: remove scaling with Astropy's"
                    "  `HDU.data = HDU.data * BSCALE + BZERO` and save a new file"
                )
            # NOTE: check for primary.data size being too large removed, since
            # data is read in chunks, so no danger of exhausting memory
            # NOTE: sanity-check for ndarray type has been removed to avoid
            # forcing eager memory load of possibly very large data array.
        elif hdu.name == "BEAMS":
            beams = hdu
        else:
            raise RuntimeError(f"Unknown HDU name {hdu.name}")
    if not primary:
        raise RuntimeError(f"No PRIMARY HDU found in fits file")
    header = primary.header
    helpers = {}
    attrs = {}
    naxes = header["NAXIS"]
    helpers["naxes"] = naxes
    dim_map = _create_dim_map(helpers, header)
    _fits_header_c_values_to_metadata(helpers, header)
    if "RESTFRQ" in header:
        helpers["restfreq"] = header["RESTFRQ"]
    if "SPECSYS" in header:
        helpers["specsys"] = header["SPECSYS"]
    t_axes = helpers["t_axes"]
    if (t_axes > 0).all():
        dir_axes = t_axes[:]
        dir_axes = dir_axes - 1
        helpers["dir_axes"] = dir_axes
        dim_map["l"] = dir_axes[0]
        dim_map["m"] = dir_axes[1]
        helpers["dim_map"] = dim_map
    else:
        raise RuntimeError("Could not find both direction axes")
    if dir_axes is not None:
        attrs["direction"] = _xds_direction_attrs_from_header(helpers, header)
    helpers["has_mask"] = False
    if compute_mask:
        # 🧠 Why the primary.data reference here is Safe (does not cause
        # an eager read of entire data array)
        # primary.data is a memory-mapped array (because fits.open(..., memmap=True)
        # is used upstream)
        # da.from_array(...) wraps this without reading it immediately
        # The actual read occurs inside:
        # .map_blocks(...).any().compute()
        # ...and that triggers blockwise loading via Dask → safe and parallel
        # 💡 Gotcha
        # What would be dangerous:
        # arr = np.isnan(primary.data).any()
        # That would pull the whole array into memory. But we're not doing that.
        data_dask = da.from_array(primary.data, chunks="auto")
        # The following code black has corner case exposure, although the guard should
        # eliminate it. But there is a cleaner, dask-y way that should work that we implement
        # next, with cautions
        # def chunk_has_nan(block):
        #     if not isinstance(block, np.ndarray) or block.size == 0:
        #         return False
        #    return np.isnan(block).any()
        # helpers["has_mask"] = data_dask.map_blocks(chunk_has_nan, dtype=bool).any().compute()
        # ✅ Option: np.isnan(data_dask).any().compute()
        # 🔒 Pros:
        # Cleaner and shorter (no custom function)
        # Handles all chunk shapes robustly — no risk of empty inputs
        # Uses Dask’s own optimized blockwise operations under the hood
        # ⚠️ Cons:
        # Might trigger more eager computation if Dask can't optimize well:
        # If chunks are misaligned or small, Dask might combine many or materialize more blocks than needed
        # Especially on large images, it could bump memory pressure slightly
        # But since we already call .compute(), we will load some block data no matter
        # what — this just changes how much and how smartly.
        # ✅ Verdict for compute_mask
        # Because this is explicitly for computing a global has-NaN flag (not building the
        # dataset), recommend:
        # helpers["has_mask"] = np.isnan(data_dask).any().compute()
        # It's concise, robust to shape edge cases, and still parallelized.
        # We can always revisit it later if perf becomes a concern — and even then,
        # it's likely a matter of tuning chunks= manually rather than the expression itself.
        #
        # This compute will normally be done in parallel
        helpers["has_mask"] = np.isnan(data_dask).any().compute()
    beam = _beam_attr_from_header(helpers, header)
    if beam != "mb":
        helpers["beam"] = beam
    if "BITPIX" in header:
        v = abs(header["BITPIX"])
        if v == 16:
            helpers["dtype"] = "int16"
        if v == 32:
            helpers["dtype"] = "float32"
        elif v == 64:
            helpers["dtype"] = "float64"
        else:
            raise RuntimeError(f'Unhandled data type {header["BITPIX"]}')
    helpers["obsdate"] = make_time_measure_dict(
        data=Time(header["DATE-OBS"], format="isot").mjd,
        units=["d"],
        scale=header["TIMESYS"],
        time_format="MJD",
    )

    # TODO complete _make_history_xds when spec has been finalized
    # attrs['history'] = _make_history_xds(header)
    return attrs, helpers, header


def _make_history_xds(header):
    # TODO complete writing history when we actually have a spec for what
    # the image history is supposed to be, since doing this now may
    # be a waste of time if the final spec turns out to be significantly
    # different from our current ad hoc history xds
    # in astropy, 3506803168 seconds corresponds to 1970-01-01T00:00:00
    history_list = list(header.get("HISTORY"))
    for i in range(len(history_list) - 1, -1, -1):
        if (i == len(history_list) - 1 and history_list[i] == "CASA END LOGTABLE") or (
            i == 0 and history_list[i] == "CASA START LOGTABLE"
        ):
            history_list.pop(i)
        elif history_list[i].startswith(">"):
            # entry continuation line
            history_list[i - 1] = history_list[i - 1] + history_list[i][1:]
            history_list.pop(i)


def _create_coords(
    helpers: dict, header: fits.header, do_sky_coords: bool
) -> xr.Dataset:
    dir_axes = helpers["dir_axes"]
    dim_map = helpers["dim_map"]
    sphr_dims = (
        [dim_map["l"], dim_map["m"]] if ("l" in dim_map) and ("m" in dim_map) else []
    )
    helpers["sphr_dims"] = sphr_dims
    coords = {}
    coords["time"] = _get_time_values(helpers)
    coords["frequency"] = _get_freq_values(helpers)
    coords["polarization"] = _get_pol_values(helpers)
    coords["velocity"] = (["frequency"], _get_velocity_values(helpers))
    if len(sphr_dims) > 0:
        for i, c in enumerate(["l", "m"]):
            idx = sphr_dims[i]
            cdelt_rad = helpers["cdelt"][idx] * u.Unit(_get_unit(helpers["cunit"][idx]))
            cdelt_rad = abs(cdelt_rad.to("rad").value)
            if c == "l":
                # l values increase to the east
                # l follows RA as far as increasing/decreasing, see AIPS Meme 27,
                # change in alpha definition three lines below Figure 2 and the first
                # of the pair of equations 10.
                cdelt_rad = -cdelt_rad
            helpers[c] = {}
            helpers[c]["cunit"] = "rad"
            helpers[c]["cdelt"] = cdelt_rad
            coords[c] = _compute_linear_world_values(
                naxis=helpers["shape"][idx],
                crpix=helpers["crpix"][idx],
                crval=0.0,
                cdelt=cdelt_rad,
            )
        if do_sky_coords:
            pick = lambda mylist: [mylist[i] for i in sphr_dims]
            my_ret = _compute_world_sph_dims(
                projection=helpers["projection"],
                shape=pick(helpers["shape"]),
                ctype=pick(helpers["ctype"]),
                crpix=pick(helpers["crpix"]),
                crval=pick(helpers["crval"]),
                cdelt=pick(helpers["cdelt"]),
                cunit=pick(helpers["cunit"]),
            )
            helpers["cunit"] = my_ret["units"]
            for j, i in enumerate(dir_axes):
                helpers["crval"][i] = my_ret["ref_val"][j]
                helpers["cdelt"][i] = my_ret["inc"][j]
            coords[my_ret["axis_name"][0]] = (["l", "m"], my_ret["value"][0])
            coords[my_ret["axis_name"][1]] = (["l", "m"], my_ret["value"][1])
            helpers["sphr_axis_names"] = tuple(my_ret["axis_name"])
    else:
        # Fourier image
        coords["u"], coords["v"] = _get_uv_values(helpers)
    coords["beam_param"] = ["major", "minor", "pa"]
    xds = xr.Dataset(coords=coords)
    return xds


def _get_time_values(helpers):
    return [helpers["obsdate"]["data"]]


def _get_pol_values(helpers):

    idx = helpers["ctype"].index("STOKES")
    if idx >= 0:
        vals = []
        crval = int(helpers["crval"][idx])
        crpix = int(helpers["crpix"][idx])
        cdelt = int(helpers["cdelt"][idx])
        stokes_start_idx = crval - cdelt * crpix
        for i in range(helpers["shape"][idx]):
            stokes_idx = (stokes_start_idx + i) * cdelt
            if 0 <= stokes_idx < len(stokes_types):
                # stokes_types provides the index-label mapping from casacore Stokes.h
                vals.append(stokes_types[stokes_idx])
            else:
                raise RuntimeError(
                    "Can't find the Stokes type using the FITS header index"
                )
        return vals
    else:
        return ["I"]


def _get_freq_values(helpers: dict) -> list:
    vals = []
    ctype = helpers["ctype"]
    if "FREQ" in ctype:
        freq_idx = ctype.index("FREQ")
        helpers["freq_axis"] = freq_idx
        vals = _compute_linear_world_values(
            naxis=helpers["shape"][freq_idx],
            crval=helpers["crval"][freq_idx],
            crpix=helpers["crpix"][freq_idx],
            cdelt=helpers["cdelt"][freq_idx],
        )
        cunit = helpers["cunit"][freq_idx]
        helpers["frequency"] = vals * u.Unit(cunit)
        return vals
    elif "VOPT" in ctype:
        if "restfreq" in helpers:
            restfreq = helpers["restfreq"] * u.Hz
        else:
            raise RuntimeError(
                "Spectral axis in FITS header is velocity, but there is "
                "no rest frequency so converting to frequency is not possible"
            )
        helpers["doppler"] = "Z"
        v_idx = ctype.index("VOPT")
        helpers["freq_idx"] = v_idx
        helpers["freq_axis"] = v_idx
        crval = helpers["crval"][v_idx]
        crpix = helpers["crpix"][v_idx]
        cdelt = helpers["cdelt"][v_idx]
        cunit = helpers["cunit"][v_idx]
        freq, vel = _freq_from_vel(
            crval, cdelt, crpix, cunit, "Z", helpers["shape"][v_idx], restfreq
        )
        helpers["velocity"] = vel["value"] * u.Unit(vel["units"])
        helpers["crval"][v_idx] = (freq["crval"] * u.Unit(freq["units"])).to(u.Hz).value
        helpers["cdelt"][v_idx] = (freq["cdelt"] * u.Unit(freq["units"])).to(u.Hz).value
        return list(freq["value"])
    else:
        return [1420e6]


def _get_velocity_values(helpers: dict) -> list:
    if "velocity" in helpers:
        return helpers["velocity"].to(u.m / u.s).value
    elif "frequency" in helpers:
        v = _compute_velocity_values(
            restfreq=helpers["restfreq"],
            freq_values=helpers["frequency"].to("Hz").value,
            doppler=helpers.get("doppler", "RADIO"),
        )
        helpers["velocity"] = v * (u.m / u.s)
        return v


# FIXME change namee, even if there is only a single beam, we make a
# multi beam array using it. If we have a beam, it will always be
# "mutltibeam" is name is redundant and confusing
def _do_multibeam(xds: xr.Dataset, imname: str) -> xr.Dataset:
    """Only run if we are sure there are multiple beams"""
    hdulist = fits.open(imname)
    for hdu in hdulist:
        header = hdu.header
        if "EXTNAME" in header and header["EXTNAME"] == "BEAMS":
            units = (
                u.Unit(header["TUNIT1"]),
                u.Unit(header["TUNIT2"]),
                u.Unit(header["TUNIT3"]),
            )
            nchan = header["NCHAN"]
            npol = header["NPOL"]
            beam_array = np.zeros([1, nchan, npol, 3])
            data = hdu.data
            hdulist.close()
            for t in data:
                beam_array[0, t[3], t[4]] = t[0:3]
            for i in (0, 1, 2):
                beam_array[:, :, :, i] = (
                    (beam_array[:, :, :, i] * units[i]).to("rad").value
                )
            return _create_beam_data_var(xds, beam_array)
    raise RuntimeError(
        "It looks like there should be a BEAMS table but no "
        "such table found in FITS file"
    )


def _add_beam(xds: xr.Dataset, helpers: dict) -> xr.Dataset:
    nchan = xds.sizes["frequency"]
    npol = xds.sizes["polarization"]
    beam_array = np.zeros([1, nchan, npol, 3])
    beam_array[0, :, :, 0] = helpers["beam"]["bmaj"]["data"]
    beam_array[0, :, :, 1] = helpers["beam"]["bmin"]["data"]
    beam_array[0, :, :, 2] = helpers["beam"]["pa"]["data"]
    return _create_beam_data_var(xds, beam_array)


def _create_beam_data_var(xds: xr.Dataset, beam_array: np.array) -> xr.Dataset:
    xdb = xr.DataArray(
        beam_array, dims=["time", "frequency", "polarization", "beam_param"]
    )
    xdb = xdb.rename("BEAM")
    xdb = xdb.assign_coords(beam_param=["major", "minor", "pa"])
    xdb.attrs["units"] = "rad"
    xds["BEAM"] = xdb
    return xds


def _get_uv_values(helpers: dict) -> tuple:
    shape = helpers["shape"]
    ctype = helpers["ctype"]
    unit = helpers["cunit"]
    delt = helpers["cdelt"]
    ref_pix = helpers["crpix"]
    ref_val = helpers["crval"]
    for i, axis in enumerate(["UU", "VV"]):
        idx = ctype.index(axis)
        if idx >= 0:
            z = []
            crpix = ref_pix[i]
            crval = ref_val[i]
            cdelt = delt[i]
            for i in range(shape[idx]):
                f = (i - crpix) * cdelt + crval
                z.append(f)
            if axis == "UU":
                u = z
            else:
                v = z
    return u, v


def _add_sky_or_aperture(
    xds: xr.Dataset,
    ary: Union[np.ndarray, da.array],
    dim_order: list,
    header,
    helpers: dict,
    has_sph_dims: bool,
) -> xr.Dataset:
    xda = xr.DataArray(ary, dims=dim_order)
    for h, a in zip(
        ["BUNIT", "BTYPE", "OBJECT", "OBSERVER"],
        ["units", _image_type, "object_name", "observer"],
    ):
        if h in header:
            xda.attrs[a] = header[h]
    xda.attrs["obsdate"] = helpers["obsdate"].copy()
    xda.attrs["pointing_center"] = _compute_pointing_center(helpers, header)
    xda.attrs["telescope"] = _get_telescope_metadata(helpers, header)
    xda.attrs["description"] = None
    xda.attrs["user"] = _user_attrs_from_header(header)
    name = "SKY" if has_sph_dims else "APERTURE"
    xda = xda.rename(name)
    xds[xda.name] = xda
    if helpers["has_mask"]:
        pp = da if type(xda[0].data) == dask.array.core.Array else np
        mask = pp.isnan(xda)
        mask.attrs = {}
        mask = mask.rename("MASK0")
        xds["MASK0"] = mask
        xda.attrs["active_mask"] = "MASK0"
    xda = xda.rename(name)
    xds[xda.name] = xda
    return xds


def _read_image_array(
    img_full_path: str, chunks: dict, helpers: dict, verbose: bool
) -> da.array:
    # memmap = True allows only part of data to be loaded into memory
    # may also need to pass mode='denywrite'
    # https://stackoverflow.com/questions/35759713/astropy-io-fits-read-row-from-large-fits-file-with-mutliple-hdus
    if isinstance(chunks, dict):
        mychunks = _get_chunk_list(chunks, helpers)
    else:
        raise ValueError(
            f"incorrect type {type(chunks)} for parameter chunks. Must be dict"
        )
    transpose_list, new_axes = _get_transpose_list(helpers)
    data_type = helpers["dtype"]
    rshape = helpers["shape"][::-1]
    full_chunks = mychunks + tuple([1 for rr in range(5) if rr >= len(mychunks)])
    d0slices = []
    blc = tuple(5 * [0])
    trc = tuple(rshape) + tuple([1 for rr in range(5) if rr >= len(mychunks)])
    for d0 in range(blc[0], trc[0], full_chunks[0]):
        d0len = min(full_chunks[0], trc[0] - d0)
        d1slices = []
        for d1 in range(blc[1], trc[1], full_chunks[1]):
            d1len = min(full_chunks[1], trc[1] - d1)
            d2slices = []
            for d2 in range(blc[2], trc[2], full_chunks[2]):
                d2len = min(full_chunks[2], trc[2] - d2)
                d3slices = []
                for d3 in range(blc[3], trc[3], full_chunks[3]):
                    d3len = min(full_chunks[3], trc[3] - d3)
                    d4slices = []
                    for d4 in range(blc[4], trc[4], full_chunks[4]):
                        d4len = min(full_chunks[4], trc[4] - d4)
                        shapes = tuple(
                            [d0len, d1len, d2len, d3len, d4len][: len(rshape)]
                        )
                        starts = tuple([d0, d1, d2, d3, d4][: len(rshape)])
                        delayed_array = dask.delayed(_read_image_chunk)(
                            img_full_path, shapes, starts
                        )
                        d4slices += [da.from_delayed(delayed_array, shapes, data_type)]
                    d3slices += (
                        [da.concatenate(d4slices, axis=4)]
                        if len(rshape) > 4
                        else d4slices
                    )
                d2slices += (
                    [da.concatenate(d3slices, axis=3)] if len(rshape) > 3 else d3slices
                )
            d1slices += (
                [da.concatenate(d2slices, axis=2)] if len(rshape) > 2 else d2slices
            )
        d0slices += [da.concatenate(d1slices, axis=1)] if len(rshape) > 1 else d1slices
    ary = da.concatenate(d0slices, axis=0)
    ary = da.expand_dims(ary, new_axes)
    return ary.transpose(transpose_list)


def _get_chunk_list(chunks: dict, helpers: dict) -> tuple:
    ret_list = list(helpers["shape"])[::-1]
    axis = 0
    ctype = helpers["ctype"]
    for c in ctype[::-1]:
        if c.startswith("RA"):
            if "l" in chunks:
                ret_list[axis] = chunks["l"]
        elif c.startswith("DEC"):
            if "m" in chunks:
                ret_list[axis] = chunks["m"]
        elif c.startswith("FREQ") or c.startswith("VOPT") or c.startswith("VRAD"):
            if "frequency" in chunks:
                ret_list[axis] = chunks["frequency"]
        elif c.startswith("STOKES"):
            if "polarization" in chunks:
                ret_list[axis] = chunks["polarization"]
        else:
            raise RuntimeError(f"Unhandled coordinate type {c}")
        axis += 1
    return tuple(ret_list)


def _get_transpose_list(helpers: dict) -> tuple:
    ctype = helpers["ctype"]
    transpose_list = 5 * [-1]
    # time axis
    transpose_list[0] = 4
    new_axes = [4]
    last_axis = 3
    not_covered = ["l", "m", "u", "v", "s", "f"]
    for i, c in enumerate(ctype[::-1]):
        b = c.lower()
        if b.startswith("ra") or b.startswith("uu"):
            transpose_list[3] = i
            not_covered.remove("l")
            not_covered.remove("u")
        elif b.startswith("dec") or b.startswith("vv"):
            transpose_list[4] = i
            not_covered.remove("m")
            not_covered.remove("v")
        elif (
            b.startswith("frequency")
            or b.startswith("freq")
            or b.startswith("vopt")
            or b.startswith("vrad")
        ):
            transpose_list[1] = i
            not_covered.remove("f")
        elif b.startswith("stok"):
            transpose_list[2] = i
            not_covered.remove("s")
        else:
            raise RuntimeError(f"Unhandled axis name {c}")
    h = {"l": 3, "m": 4, "u": 3, "v": 4, "f": 2, "s": 1}
    for p in not_covered:
        transpose_list[h[p]] = last_axis
        new_axes.append(last_axis)
        last_axis -= 1
    new_axes.sort()
    if transpose_list.count(-1) > 0:
        raise RuntimeError(f"Logic error: axes {axes}, transpose_list {transpose_list}")
    return transpose_list, new_axes


def _read_image_chunk(img_full_path, shapes: tuple, starts: tuple) -> np.ndarray:
    hdulist = fits.open(img_full_path, memmap=True)
    hdu = hdulist[0]
    # Chunk slice
    slices = tuple(
        slice(start, start + length) for start, length in zip(starts, shapes)
    )
    chunk = hdu.data[slices]
    hdulist.close()
    del hdulist
    return chunk
